/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2008 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: simplest.c,v 1.90 2007/09/27 07:37:20 zlb Exp $ */

#include "phg.h"
#include <string.h>
#include <math.h>

static void
func_f(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
    FLOAT tmp = -10.0 * (x * x + y * y + z * z);
    *value = 40.0 * (1.5 + tmp) * exp(tmp);
}

static void
func_u(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
    *value = exp(-10.0 * (x * x + y * y + z * z));
}

static void
build_linear_system(SOLVER *solver, DOF *u_h, DOF *f_h)
{
    int N = u_h->type->nbas;	/* number of basis functions in an element */
    int i, j;
    GRID *g = u_h->g;
    SIMPLEX *e;
    FLOAT A[N][N], rhs[N], buffer[N];
    INT I[N];

    assert(u_h->dim == 1);
    ForAllElements(g, e) {
	/* compute \int \grad\phi_i \cdot \grad\phi_j making use of symmetry */
	for (i = 0; i < N; i++) {
	    I[i] = phgSolverMapE2L(solver, 0, e, i);
	    for (j = 0; j <= i; j++)
		A[j][i] = A[i][j] =
		    phgQuadGradBasDotGradBas(e, u_h, i, u_h, j, QUAD_DEFAULT);
	}

	/* loop on basis functions */
	for (i = 0; i < N; i++) {
	    if (phgDofDirichletBC(u_h, e, i, func_u, buffer, rhs+i,
					DOF_PROJ_NONE)) {
		phgSolverAddMatrixEntries(solver, 1, I + i, N, I, buffer); 
	    }
	    else {	/* interior node */
		/* right hand side = \int f * phi_i */
		phgQuadDofTimesBas(e, f_h, u_h, i, QUAD_DEFAULT, rhs + i);
		phgSolverAddMatrixEntries(solver, 1, I + i, N, I, A[i]); 
	    }
	}
	phgSolverAddRHSEntries(solver, N, I, rhs);
    }
}

static FLOAT
estimate_error(DOF *u_h, DOF *f_h, DOF *grad_u, DOF *error)
/* compute H1 error indicator */
{
    GRID *g = u_h->g;
    SIMPLEX *e;
    DOF *jump, *residual, *tmp;

    jump = phgQuadFaceJump(grad_u, DOF_PROJ_DOT, NULL, QUAD_DEFAULT);
    tmp = phgDofDivergence(grad_u, NULL, NULL, NULL);
    residual = phgDofCopy(f_h, NULL, DOF_DGn[u_h->type->order], NULL); 
    phgDofAXPY(1.0, tmp, &residual);
    phgDofFree(&tmp);
    ForAllElements(g, e) {
	int i;
	FLOAT eta, h;
	FLOAT diam = phgGeomGetDiameter(g, e);
	eta = 0.0;
	/* for each face F compute [grad_u \cdot n] */
	for (i = 0; i < NFace; i++) {
	    if (e->bound_type[i] & (DIRICHLET | NEUMANN))
		continue;	/* boundary face */
	    h = phgGeomGetFaceDiameter(g, e, i);
	    eta += *DofFaceData(jump, e->faces[i]) * h;
	}
	eta = eta * .5 + diam * diam * phgQuadDofDotDof(e, residual, residual,
								QUAD_DEFAULT);
	*DofElementData(error, e->index) = Sqrt(eta);
    }
    phgDofFree(&jump);
    phgDofFree(&residual);
    return phgDofNormInftyVec(error);
}

int
main(int argc, char *argv[])
{
    GRID *g;
    SIMPLEX *e;
    DOF *u_h, *grad_u, *f_h, *error;
    SOLVER *solver;
    FLOAT PE_error = 0.0, thres;
    char *fn = "../test/cube.dat";
    size_t mem, mem_peak;

    phgInit(&argc, &argv);
    g = phgNewGrid(-1);
    if (!phgImport(g, fn, FALSE))
	phgError(1, "can't read file \"%s\".\n", fn);
    u_h = phgDofNew(g, DOF_DEFAULT, 1, "u_h", DofInterpolation);
    phgDofSetDataByValue(u_h, 0.0);
    f_h = phgDofNew(g, DOF_DEFAULT, 1, "f_h",  func_f);
    error = phgDofNew(g, DOF_P0, 1, "error indicator", DofNoAction);

    while (TRUE) {
	phgPrintf("%d DOF, %d elements, %d submeshes, load imbalance: %lg\n",
			DofGetDataCountGlobal(u_h), g->nleaf_global, g->nprocs,
			(double)g->lif);
	if (phgBalanceGrid(g, 1.2, -1, NULL, 0.))
	    phgPrintf("Repartition mesh, load imbalance: %lg\n",
			(double)g->lif);
	solver = phgSolverCreate(SOLVER_DEFAULT, u_h, NULL);
	build_linear_system(solver, u_h, f_h);
	phgSolverSolve(solver, TRUE, u_h, NULL);
	phgSolverDestroy(&solver);
	grad_u = phgDofGradient(u_h, NULL, NULL, NULL);
	PE_error = estimate_error(u_h, f_h, grad_u, error);
	mem = phgMemoryUsage(g, &mem_peak);
	phgPrintf("  Error indicator = %0.3le, memory = %0.2lfMB\n",
			(double)PE_error, (double)mem_peak / (1024.0 * 1024.0));
	phgDofFree(&grad_u);
	if (PE_error < 1e-3 || mem_peak >= 1024 * (size_t)400 * 1024)
	    break;
	thres = 1e-3 / Sqrt((double)g->nleaf_global);
	ForAllElements(g, e)
	    e->mark = (*DofElementData(error, e->index) > thres) ? 1 : 0;
	phgRefineMarkedElements(g);
    }
#if 0
    phgPrintf("Final mesh written to \"%s\".\n",
	phgExportVTK(g, "output.vtk", u_h, error, NULL));
#endif
    phgDofFree(&u_h);
    phgDofFree(&f_h);
    phgDofFree(&error);
    phgFreeGrid(&g);
    phgFinalize();
    return 0;
}
