/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2008 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: mark.h,v 1.5 2008/01/16 04:24:06 liuhui Exp $ */

#ifndef PHG_MARK_H

/******************************************************************************
 * DOF *error and DOF *osc must have one value on each element or each face
 * which is the (L^p norm)^p of the error on the element or face
 * (normally p = 2)
 *                                                                 
 * The type of error indicator defined over element 
 * must be DOF_P0                                                 
 * the type of error indicator defined over face                   
 * must be the same as the type of Face Jump                       
 * the type of data oscillation must be DOF_P0 and                 
 * defined over element                                            
 * ***************************************************************************/

/* Strategies */
enum {
     MARK_DEFAULT,                /* use default strategy */
     MARK_NONE,                        /* don't mark any elements */
     MARK_ALL,                        /* mark all elements (uniformly refine) */
     MARK_MAX,                        /* MAX strategy */
     MARK_GERS,                        /* GERS strategy */
     MARK_MNS,                        /* MNS strategy */
     MARK_EQDIST        /* equidistribution strategy */
};

#ifdef __cplusplus
extern "C" {
#endif

/******************************************************************
 *                  
 *                  Marking function
 *
 *
 *   'theta' >= 0.0, 'zeta' >= 0., 'gamma' >= 0.
 *   'p' >= 1.0, 'coarsen_p' >= 1.
 *   'depth' >= 1, int
 *   'caorsen_depth' >= 1, int
 *
 *   'error', 'osc' and 'coarsen_error' should contain elementwise or facewise
 *   errors to the power of 'p' ('coarsen_p' for 'coarsen_error'),
 *   'error' cannot be NULL except for the first call (initialization),
 *   or the strategy is mark_all or mark_none.
 *   'osc' and 'coarsen_error' may be NULL
 *   coarsen_error can't be NULL if coarsen_strategy is not mark_none
 *
 *   Options for -default_strategy: max, gers, mns, all, equidistribution
 *        default: gers
 *   
 *   It's not necessary to specify the types of DOFs,
 *        the subroutine can probe them, but the strategy and
 *        the types of DOFs must match
 *   DOF of coarsening error should contain element data.
 *                  
 *****************************************************************/

void phgMarkElements_(
        int strategy, DOF *error, FLOAT theta, DOF *osc, FLOAT zeta,
                int depth, FLOAT p,
        int coarsen_strategy, DOF *coarsen_error, FLOAT gamma,
                int coarsen_depth, FLOAT coarsen_p, FLOAT tol);

/* refine && coarsen */
#define phgMarkElements(strategy, error, theta, osc, zeta, depth, p,\
        coarsen_strategy, coarsen_error, gamma, coarsen_depth,\
        coarsen_p, tol)  \
    phgMarkElements_(strategy, error, theta, osc, zeta, depth, p,\
        coarsen_strategy, coarsen_error, gamma, coarsen_depth,\
        coarsen_p, tol)

/* refine only */
#define phgMarkRefine(strategy, error, theta, osc, zeta, depth, p, tol) \
    phgMarkElements_(strategy, error, theta, osc, zeta, depth, p,\
                     MARK_NONE, NULL, 0.5, 1, 2, tol)

/* coarsen only */
#define phgMarkCoarsen(strategy, error, gama, depth, p, tol) \
    phgMarkElements_(MARK_NONE, NULL, 0.5, NULL, 0.5, 1, 2,\
             strategy, error, gama, depth, p, tol)

#ifdef __cplusplus
}
#endif

#define PHG_MARK_H
#endif
