/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2008 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: coarsen.c,v 1.152 2008/01/13 09:42:48 zlb Exp $
 *
 * The serial algorithm:
 *
 * On entry, for any leaf element e, -(e->mark) is the max number of times e
 * is allowed to be coarsened.
 * 
 *   1. The mesh is traversed using function set_mark(), in which
 *	    e->mark = max(e->children[0]->mark, e->children[1]->mark) + 1
 *	is computed for all non leaf elements, e->flag is set to 0 for all
 *	elements, neighbour links and boundary types are updated for non
 *	leaf elements.
 *
 *   2. check_patch() is called to check all patches around the refinement
 *	edges of non leaf elements. If one of the element on a patch has
 *	e->mark > 0, then the patch can't be coarsened, e->flag is recursively
 *	set to CHECK_FAILED for all elements in the patch, as well as those
 *	on the refinement patches of their ancestors.
 *
 *   3. On each branch, prune the children of the top-most element having
 *	e->flag != CHECK_FAILED && e->mark <= 0.
 *
 * The above algorithm is proved to produce a conforming mesh.
 */

#include "phg.h"
#include <stdlib.h>
#include <string.h>

#define ONE_LEVEL_COARSENING	0 /* level by level coarsening if nprocs > 1 */

#define CHECK_FAILED	1
#define CHECK_PASSED	2
#define COLLECTED	3
#define REMOVED		255

/* Note: GetNeighbou() relies on the REMOTE bit which is not usable for non
 * leaf elements */
#define GetNeighbour0(e, v) (IsLeaf(e) ? GetNeighbour(e, v) : e->neighbours[v])

static GRID *g_;

static void
set_mark(GRID *g, SIMPLEX *e)
/* computes e->mark for non leaf elements, sets e->flag=0 for all elements */
{
    SIMPLEX *e0, *e1;
    SHORT mark;

    if (e == NULL)
	return;

    e->flag = 0;

    e0 = e->children[0];
    e1 = e->children[1];

    if (e0 == NULL && e1 == NULL)
	return;

    set_mark(g, e0);
    set_mark(g, e1);

    if (e0 != NULL) {
	e->mark = e0->mark + 1;
	if (e1 != NULL) {
	    e->mark = (e->mark >= (mark = e1->mark + 1)) ? e->mark : mark;
	}
    }
    else {	/* e1 != NULL */
	e->mark = e1->mark + 1;
    }

    return;
}

static void
free_children(GRID *g, SIMPLEX *e)
{
#if ONE_LEVEL_COARSENING
    SHORT mark;
#endif	/* ONE_LEVEL_COARSENING */
    SIMPLEX *e0 = e->children[0], *e1 = e->children[1];

    if (e0 == NULL && e1 == NULL)
	return;

    if (e0 != NULL)
	free_children(g, e0);

    if (e1 != NULL)
	free_children(g, e1);

#if ONE_LEVEL_COARSENING
    /* update e->mark */
    if (e0 != NULL) {
	e->mark = e0->mark + 1;
	if (e1 != NULL) {
	    e->mark = (e->mark >= (mark = e1->mark + 1)) ? e->mark : mark;
	}
    }
    else {	/* e1 != NULL */
	e->mark = e1->mark + 1;
    }
#endif	/* ONE_LEVEL_COARSENING */

    if (e0 != NULL) {
	phgFreeElement(&e0);
	e->children[0] = NULL;
    }

    if (e1 != NULL) {
	phgFreeElement(&e1);
	e->children[1] = NULL;
    }
}

#if DEBUG
static int
get_patch(GRID *g,  SIMPLEX *e, SIMPLEX ***p)
/* returns the patch composed of all elements having edge 0 of 'e' as their
 * edge 0. Note: phgGetPatch() here requires neighbours links of non-leaf
 * elements, which are updated by the function set_neighbours() */
{
    int i, count;
    INT v0, v1;

    v0 = e->verts[0];
    v1 = e->verts[1];
    count = phgGetPatch(g, e, 0, 1, p);
    for (i = 0; i < count; i++) {
	e = (*p)[i];
	assert((v0 == e->verts[0] && v1 == e->verts[1])
		|| (v0 == e->verts[1] && v1 == e->verts[0]));
    }

    return count;
}
#else	/* DEBUG */
# define get_patch(g, e, p)	phgGetPatch(g, e, 0, 1, p)
#endif	/* DEBUG */

static void
set_uncoarsen(GRID *g, SIMPLEX *e)
{
    SIMPLEX **p;
    int i, count;

    if (e == NULL || e->flag == CHECK_FAILED)
	return;

    count = get_patch(g, e, &p);
    for (i = 0; i < count; i++)
	p[i]->flag = CHECK_FAILED;
    for (i = 0; i < count; i++)
	set_uncoarsen(g, p[i]->parent);
    phgFree(p);
}

static void
check_patch(GRID *g, SIMPLEX *e)
{
    SIMPLEX **p;
    int i, count;

    if (e == NULL || IsLeaf(e))
	return;

    check_patch(g, e->children[0]);
    check_patch(g, e->children[1]);

    if (e->flag == CHECK_FAILED || e->flag == CHECK_PASSED)
	return;

    count = get_patch(g, e, &p);
    for (i = 0; i < count; i++) {
	if (p[i]->mark > 0)
	    break;
	e->flag = CHECK_PASSED;
    }

    if (i >= count) {
	/* all elements on the patch passed the test */
	phgFree(p);
	return;
    }

    /* the patch can't be coarsened, set e->flag == CHECK_FAILED for all
     * elements on the patch and all their ancestors */
    for (i = 0; i < count; i++)
	p[i]->flag = CHECK_FAILED;
    for (i = 0; i < count; i++)
	set_uncoarsen(g, p[i]->parent);
    phgFree(p);
}

static void
do_coarsen(GRID *g, SIMPLEX *e, INT *ntree, INT *nleaf)
{
    SIMPLEX *ee;

    if (e == NULL)
	return;

    if (e->flag != REMOVED)
	(*ntree)++;

    if ((ee = e->children[0]) != NULL && ee->flag == REMOVED) {
	free_children(g, ee);
	phgFreeElement(&ee);
	e->children[0] = NULL;
    }

    if ((ee = e->children[1]) != NULL && ee->flag == REMOVED) {
	free_children(g, ee);
	phgFreeElement(&ee);
	e->children[1] = NULL;
    }
    if ((e->mark <= 0 && e->flag != CHECK_FAILED) || IsLeaf(e)) {
	/* prune children of e */
	free_children(g, e);
	if (e->flag != REMOVED)	/* this test is necessary for root elements */
	    (*nleaf)++;
	return;
    }

    do_coarsen(g, e->children[0], ntree, nleaf);
    do_coarsen(g, e->children[1], ntree, nleaf);

    return;
}

static INT *map;
static size_t map_local_offset, map_local_size, map_local_max;

static BOOLEAN
map_collect_cb CB_ARGS(e)
{
    int i;
    INT *local_map = (INT *)(((BYTE *)e) + map_local_offset);

    /* make phgUpdateBoundaryTypes() recompute ordering */
    e->ordering = 0;

    for (i = 0; i < map_local_size; i++) {
	assert(local_map[i] >=0 && local_map[i] < map_local_max);
	map[local_map[i]] = 1;
    }

    return TRUE;
}

static BOOLEAN
map_update_cb CB_ARGS(e)
{
    int i;
    INT *local_map = (INT *)(((BYTE *)e) + map_local_offset);

    for (i = 0; i < map_local_size; i++) {
	local_map[i] = map[local_map[i]];
	assert(local_map[i] >= 0);
    }

    return TRUE;
}

static BOOLEAN dof_interp_flag = FALSE;	/* indicates whether interpolate DOFs */

static BOOLEAN
pack_indices(GRID *g, GTYPE type)
/* updates indices and L2Gmap. `type' may be VERTEX, EDGE, FACE, or ELEMENT */
{
    SIMPLEX dummy;
    void (*traverse)(GRID *g, BOOLEAN (*cb) CB_ARGS(e)) = NULL;
    INT i, j, newcount, *count = NULL, *count_global = NULL;

    switch (type) {
	case VERTEX:
	    traverse = phgTraverseAllElements;
	    count = &g->nvert;
	    count_global = &g->nvert_global;
	    map_local_offset = (BYTE *)(dummy.verts) - ((BYTE *)&dummy);
	    map_local_size = NVert;
	    map_local_max = g->nvert;
	    break;
	case EDGE:
	    traverse = phgTraverseAllElements;
	    count = &g->nedge;
	    count_global = &g->nedge_global;
	    map_local_offset = (BYTE *)(dummy.edges) - ((BYTE *)&dummy);
	    map_local_size = NEdge;
	    map_local_max = g->nedge;
	    break;
	case FACE:
	    traverse = phgTraverseAllElements;
	    count = &g->nface;
	    count_global = &g->nface_global;
	    map_local_offset = (BYTE *)(dummy.faces) - ((BYTE *)&dummy);
	    map_local_size = NFace;
	    map_local_max = g->nface;
	    break;
	case ELEMENT:
	    traverse = phgTraverseAllElements;
	    count = &g->nelem;
	    count_global = &g->nelem_global;
	    map_local_offset = (BYTE *)(&dummy.index) - ((BYTE *)&dummy);
	    map_local_size = 1;
	    map_local_max = g->nelem;
	    break;
    }

    /* collect referenced indices */
    map = phgCalloc(map_local_max, sizeof(*map));
    (*traverse)(g, map_collect_cb);

    newcount = -1;
    for (i = 0; i < *count; i++) {
	if (map[i] != 0)
	    map[i] = ++newcount;
	else
	    map[i] = newcount;
    }
    newcount++;

    /* pack DOF values */
    if (dof_interp_flag == TRUE) {
	DOF *dof, **pdof;
	FLOAT *data0 = NULL, *data1 = NULL;
	size_t size = 0;

	for (i = 0; i < *count; i++) {
	    if (map[i] < 0 || (i > 0 && map[i] == map[i - 1]))
		continue;
	    for (pdof = g->dof; (dof = *pdof) != NULL; pdof++) {
		if (dof->data == NULL ||
		    SpecialDofType(dof->type) ||
		    dof->type->InterpF2C == NULL ||
		    dof->userfunc == DofNoAction)
		    continue;
		/* Note: at this stage dof->data_{vert,edge,face,elem} still
		 * point to old locations, bug g->{nvert,nedge,nface,nelem}
		 * are updated in order (e.g., if type == FACE, then g->nvert
		 * and g->nedge contain new values) */
		switch (type) {
		    case VERTEX:
			size = dof->count_vert;
			data0 = dof->data + map[i] * size;
			data1 = DofVertexData(dof, i);
			break;
		    case EDGE:
			size = dof->count_edge;
			data0 = dof->data + g->nvert * dof->count_vert
				+ map[i] * size;
			data1 = DofEdgeData(dof, i);
			break;
		    case FACE:
			size = dof->count_face;
			data0 = dof->data + g->nvert * dof->count_vert
					  + g->nedge * dof->count_edge
				+ map[i] * size;
			data1 = DofFaceData(dof, i);
			break;
		    case ELEMENT:
			size = dof->count_elem;
			data0 = dof->data + g->nvert * dof->count_vert
					  + g->nedge * dof->count_edge
					  + g->nface * dof->count_face
				+ map[i] * size;
			data1 = DofElementData(dof, i);
			break;
		}
		if (size == 0 || data0 == data1)
		    continue;
		memcpy(data0, data1, size * sizeof(FLOAT));
	    }
	}
    }

    if (newcount == *count) {
	/* no change */
	phgFree(map);
	return FALSE;
    }

    /* assign new indices */
    (*traverse)(g, map_update_cb);

    if (type == VERTEX) {
	/* update g->verts[] */
	COORD *verts;
	verts = phgAlloc(newcount * sizeof(*verts));
	for (i = j = 0; i < *count; i++) {
	    if (map[i] >= 0 && (i == 0 || map[i] != map[i - 1]))
		memcpy(verts + (j++), g->verts + i, sizeof(*verts));
	}
	phgFree(g->verts);
	g->verts = verts;
    }

    *count = newcount;
    if (g->nprocs == 1)
	*count_global = newcount;

    phgFree(map);
    return TRUE;
}

static struct {
    SIMPLEX	*e0, *e1;
    BYTE	v0, v1;
} *faces, *pf;

static BOOLEAN
update_neighbours_callback CB_ARGS(e)
{
    int v;

    for (v = 0; v < NFace; v++) {
	pf = faces + e->faces[v];
	if (pf->e0 == NULL) {
	    pf->e0 = e;
	    pf->v0 = v;
	}
	else {
	    if (pf->e1 != NULL) {
		phgDumpElement(g_, pf->e0);
		phgDumpElement(g_, pf->e1);
		phgDumpElement(g_, e);
		phgError(0, "%s:%d, duplicate element.\n", __FILE__, __LINE__);
	    }
	    pf->e1 = e;
	    pf->v1 = v;
	}
	/* clear neighbour */
	e->neighbours[v] = NULL;
    }

    return TRUE;
}

static void
update_neighbours(GRID *g)
{
    SIMPLEX *e0, *e1;
    INT i;
    int v0, v1;

    if (!g->face_flag) {
	phgUpdateNeighbours(g);
	return;
    }

    faces = phgCalloc(g->nface, sizeof(*faces));
    phgTraverseElements(g, update_neighbours_callback);
    for (i = 0, pf = faces; i < g->nface; i++, pf++) {
	if ((e0 = pf->e0) == NULL)
	    continue;
	if ((e1 = pf->e1) == NULL)
	    continue;
	e0->bound_type[v0 = pf->v0] &= ~REMOTE;
	e0->neighbours[v0] = e1;
	e1->bound_type[v1 = pf->v1] &= ~REMOTE;
	e1->neighbours[v1] = e0;
    }
    phgFree(faces);

    return;
}

/*----------------------- DOF interpolation ------------------------*/

static DOF *dof;
static FLOAT *dof_data_new;
static BYTE *flags;
static int pass;	/* only used when userfunc != DofInterpolation */

static void
dof_interp_recursive(SIMPLEX *e, FLOAT **parent_data)
{
    GRID *g = dof->g;
    SIMPLEX *e0 = e->children[0], *e1 = e->children[1];
    DOF_USER_FUNC f = dof->userfunc;
    FLOAT *p, *tmp, *data[15];
    BYTE *pf = flags;
    INT i;
    int vmap[NVert], emap[NEdge];
    size_t size;
    static int level = 0;

    if (e0 == NULL && e1 == NULL) {
	/* leaf element, simply copy data */
	if (f != DofInterpolation && pass > 0)
	    return;
	if (level > 0) {
	    if (dof->count_vert > 0) {
		size = dof->count_vert * sizeof(FLOAT);
		memcpy(parent_data[0], DofVertexData(dof, e->verts[0]), size);
		memcpy(parent_data[1], DofVertexData(dof, e->verts[1]), size);
		memcpy(parent_data[2], DofVertexData(dof, e->verts[2]), size);
		memcpy(parent_data[3], DofVertexData(dof, e->verts[3]), size);
	    }
	    if (dof->count_edge > 0) {
		size = dof->count_edge * sizeof(FLOAT);
		memcpy(parent_data[4], DofEdgeData(dof, e->edges[0]), size);
		memcpy(parent_data[5], DofEdgeData(dof, e->edges[1]), size);
		memcpy(parent_data[6], DofEdgeData(dof, e->edges[2]), size);
		memcpy(parent_data[7], DofEdgeData(dof, e->edges[3]), size);
		memcpy(parent_data[8], DofEdgeData(dof, e->edges[4]), size);
		memcpy(parent_data[9], DofEdgeData(dof, e->edges[5]), size);
	    }
	    if (dof->count_face > 0) {
		size = dof->count_face * sizeof(FLOAT);
		memcpy(parent_data[10], DofFaceData(dof, e->faces[0]), size);
		memcpy(parent_data[11], DofFaceData(dof, e->faces[1]), size);
		memcpy(parent_data[12], DofFaceData(dof, e->faces[2]), size);
		memcpy(parent_data[13], DofFaceData(dof, e->faces[3]), size);
	    }
	}
	else {
	    if (dof->count_vert > 0) {
		size = dof->count_vert * sizeof(FLOAT);
		if (pf[i = e->verts[0]] == 0) {
		    pf[i] = 1;
		    memcpy(parent_data[0], DofVertexData(dof, i), size);
		}
		if (pf[i = e->verts[1]] == 0) {
		    pf[i] = 1;
		    memcpy(parent_data[1], DofVertexData(dof, i), size);
		}
		if (pf[i = e->verts[2]] == 0) {
		    pf[i] = 1;
		    memcpy(parent_data[2], DofVertexData(dof, i), size);
		}
		if (pf[i = e->verts[3]] == 0) {
		    pf[i] = 1;
		    memcpy(parent_data[3], DofVertexData(dof, i), size);
		}
		pf += g->nvert;
	    }
	    if (dof->count_edge > 0) {
		size = dof->count_edge * sizeof(FLOAT);
		if (pf[i = e->edges[0]] == 0) {
		    pf[i] = 1;
		    memcpy(parent_data[4], DofEdgeData(dof, i), size);
		}
		if (pf[i = e->edges[1]] == 0) {
		    pf[i] = 1;
		    memcpy(parent_data[5], DofEdgeData(dof, i), size);
		}
		if (pf[i = e->edges[2]] == 0) {
		    pf[i] = 1;
		    memcpy(parent_data[6], DofEdgeData(dof, i), size);
		}
		if (pf[i = e->edges[3]] == 0) {
		    pf[i] = 1;
		    memcpy(parent_data[7], DofEdgeData(dof, i), size);
		}
		if (pf[i = e->edges[4]] == 0) {
		    pf[i] = 1;
		    memcpy(parent_data[8], DofEdgeData(dof, i), size);
		}
		if (pf[i = e->edges[5]] == 0) {
		    pf[i] = 1;
		    memcpy(parent_data[9], DofEdgeData(dof, i), size);
		}
		pf += g->nedge;
	    }
	    if (dof->count_face > 0) {
		size = dof->count_face * sizeof(FLOAT);
		if (pf[i = e->faces[0]] == 0) {
		    pf[i] = 1;
		    memcpy(parent_data[10], DofFaceData(dof, i), size);
		}
		if (pf[i = e->faces[1]] == 0) {
		    pf[i] = 1;
		    memcpy(parent_data[11], DofFaceData(dof, i), size);
		}
		if (pf[i = e->faces[2]] == 0) {
		    pf[i] = 1;
		    memcpy(parent_data[12], DofFaceData(dof, i), size);
		}
		if (pf[i = e->faces[3]] == 0) {
		    pf[i] = 1;
		    memcpy(parent_data[13], DofFaceData(dof, i), size);
		}
	    }
	}
	if (dof->count_elem > 0) {
	    size = dof->count_elem * sizeof(FLOAT);
	    memcpy(parent_data[14], DofElementData(dof, e->index), size);
	}
	return;
    }

    if (f != DofInterpolation) {
	if (pass == 0)
	    return;

	/* recompute values on new elements */
	if (dof->count_vert > 0) {
	    if (pf[i = e->verts[0]] == 0) {
		pf[i] = 1;
		dof->type->InitFunc(dof, e, VERTEX, 0, f, NULL, NULL,
					parent_data[0], parent_data);
	    }
	    if (pf[i = e->verts[1]] == 0) {
		pf[i] = 1;
		dof->type->InitFunc(dof, e, VERTEX, 1, f, NULL, NULL,
					parent_data[1], parent_data);
	    }
	    if (pf[i = e->verts[2]] == 0) {
		pf[i] = 1;
		dof->type->InitFunc(dof, e, VERTEX, 2, f, NULL, NULL,
					parent_data[2], parent_data);
	    }
	    if (pf[i = e->verts[3]] == 0) {
		pf[i] = 1;
		dof->type->InitFunc(dof, e, VERTEX, 3, f, NULL, NULL,
					parent_data[3], parent_data);
	    }
	    pf += g->nvert;
	}
	if (dof->count_edge > 0) {
	    if (pf[i = e->edges[0]] == 0) {
		pf[i] = 1;
		dof->type->InitFunc(dof, e, EDGE, 0, f, NULL, NULL,
					parent_data[4 + 0], parent_data);
	    }
	    if (pf[i = e->edges[1]] == 0) {
		pf[i] = 1;
		dof->type->InitFunc(dof, e, EDGE, 1, f, NULL, NULL,
					parent_data[4 + 1], parent_data);
	    }
	    if (pf[i = e->edges[2]] == 0) {
		pf[i] = 1;
		dof->type->InitFunc(dof, e, EDGE, 2, f, NULL, NULL,
					parent_data[4 + 2], parent_data);
	    }
	    if (pf[i = e->edges[3]] == 0) {
		pf[i] = 1;
		dof->type->InitFunc(dof, e, EDGE, 3, f, NULL, NULL,
					parent_data[4 + 3], parent_data);
	    }
	    if (pf[i = e->edges[4]] == 0) {
		pf[i] = 1;
		dof->type->InitFunc(dof, e, EDGE, 4, f, NULL, NULL,
					parent_data[4 + 4], parent_data);
	    }
	    if (pf[i = e->edges[5]] == 0) {
		pf[i] = 1;
		dof->type->InitFunc(dof, e, EDGE, 5, f, NULL, NULL,
					parent_data[4 + 5], parent_data);
	    }
	    pf += g->nedge;
	}
	if (dof->count_face > 0) {
	    if (pf[i = e->faces[0]] == 0) {
		pf[i] = 1;
		dof->type->InitFunc(dof, e, FACE, 0, f, NULL, NULL,
					parent_data[10 + 0], parent_data);
	    }
	    if (pf[i = e->faces[1]] == 0) {
		pf[i] = 1;
		dof->type->InitFunc(dof, e, FACE, 1, f, NULL, NULL,
					parent_data[10 + 1], parent_data);
	    }
	    if (pf[i = e->faces[2]] == 0) {
		pf[i] = 1;
		dof->type->InitFunc(dof, e, FACE, 2, f, NULL, NULL,
					parent_data[10 + 2], parent_data);
	    }
	    if (pf[i = e->faces[3]] == 0) {
		pf[i] = 1;
		dof->type->InitFunc(dof, e, FACE, 3, f, NULL, NULL,
					parent_data[10 + 3], parent_data);
	    }
	}
	if (dof->count_elem > 0) {
	    dof->type->InitFunc(dof, e, ELEMENT, 0, f, NULL, NULL,
					parent_data[14], parent_data);
	}
	return;
    }

    /* allocate buffer for storing data for e0 and e1 in the following order:
     *		1 new vertex (vn),
     *		2 cut edges (v0-vn and v1-vn),
     *		2 new edges (v2-vn and v3-vn),
     *		2 cut faces on f2 (v0-v3-vn and v1-v3-vn),
     *		2 cut faces on f3 (v0-v2-vn and v1-v2-vn),
     *		1 new face,
     *		2 new elements (e0 and e1) */
    size = 1 * dof->count_vert + 4 * dof->count_edge + 5 * dof->count_face +
	   2 * dof->count_elem;
    tmp = phgCalloc(size, sizeof(FLOAT));

    if (e0 != NULL) {
	/* build parent_data[] pointers (pointing to tmp) for e0 */
	phgMapP2C(e, vmap, emap, 0);
	p = tmp;
	if ((size = dof->count_vert) > 0) {
	    /* vertex data */
	    data[vmap[0]] = parent_data[0];
	    data[vmap[2]] = parent_data[2];
	    data[vmap[3]] = parent_data[3];
	    data[3] = p;
	    p += 1 * size;
	}
	if ((size = dof->count_edge) > 0) {
	    /* edge data */
	    data[4 + emap[1]] = parent_data[4 + 1];
	    data[4 + emap[2]] = parent_data[4 + 2];
	    data[4 + emap[5]] = parent_data[4 + 5];
	    data[4 + GetEdgeNo(3, vmap[0])] = p + 0 * size;
	    data[4 + GetEdgeNo(3, vmap[2])] = p + 2 * size;
	    data[4 + GetEdgeNo(3, vmap[3])] = p + 3 * size;
	    p += 4 * size;
	}
	if ((size = dof->count_face) > 0) {
	    /* face data */
	    data[10 + 3] = parent_data[10 + 1];
	    data[10 + vmap[2]] = p + 0 * size;
	    data[10 + vmap[3]] = p + 2 * size;
	    data[10 + vmap[0]] = p + 4 * size;
	    p += 5 * size;
	}
	if ((size = dof->count_elem) > 0) {
	    /* element data */
	    data[14] = p + 0 * size;
	}
	level++;
	dof_interp_recursive(e0, data);
	level--;
    }

    if (e1 != NULL) {
	/* build parent_data[] pointers (pointing to tmp) for e1 */
	phgMapP2C(e, vmap, emap, 1);
	p = tmp;
	if ((size = dof->count_vert) > 0) {
	    /* vertex data */
	    data[vmap[1]] = parent_data[1];
	    data[vmap[2]] = parent_data[2];
	    data[vmap[3]] = parent_data[3];
	    data[3] = p;
	    p += 1 * size;
	}
	if ((size = dof->count_edge) > 0) {
	    /* edge data */
	    data[4 + emap[3]] = parent_data[4 + 3];
	    data[4 + emap[4]] = parent_data[4 + 4];
	    data[4 + emap[5]] = parent_data[4 + 5];
	    data[4 + GetEdgeNo(3, vmap[1])] = p + 1 * size;
	    data[4 + GetEdgeNo(3, vmap[2])] = p + 2 * size;
	    data[4 + GetEdgeNo(3, vmap[3])] = p + 3 * size;
	    p += 4 * size;
	}
	if ((size = dof->count_face) > 0) {
	    /* face data */
	    data[10 + 3] = parent_data[10 + 0];	/* old face */
	    data[10 + vmap[2]] = p + 1 * size;
	    data[10 + vmap[3]] = p + 3 * size;
	    data[10 + vmap[1]] = p + 4 * size;
	    p += 5 * size;
	}
	if ((size = dof->count_elem) > 0) {
	    /* element data */
	    data[14] = p + 1 * size;
	}
	level++;
	dof_interp_recursive(e1, data);
	level--;
    }

    /* build children_data[] pointers */
    p = tmp;
    data[ 0] = p;   p += dof->count_vert;
    data[ 1] = p;   p += dof->count_edge;
    data[ 2] = p;   p += dof->count_edge;
    data[ 3] = p;   p += dof->count_edge;
    data[ 4] = p;   p += dof->count_edge;
    data[ 5] = p;   p += dof->count_face;
    data[ 6] = p;   p += dof->count_face;
    data[ 7] = p;   p += dof->count_face;
    data[ 8] = p;   p += dof->count_face;
    data[ 9] = p;   p += dof->count_face;
    data[10] = p;   p += dof->count_elem;
    data[11] = p;
    /* interpolation e->children ==> e */
    dof->type->InterpF2C(dof, e, parent_data, data);

    phgFree(tmp);
}

static void
dof_interp0(SIMPLEX *e)
{
    SIMPLEX *e0 = e->children[0], *e1 = e->children[1];
    static FLOAT *parent_data[15], *p;

    if (e->flag == REMOVED) {
	/* Note: removing this test leads to interpolation on non-leaf
	 * elements of the new grid, producing incorrect results.
	 * Test case:
	 *    mpirun -np 2 coarsen_test \
	 *		-mesh_file ../netgen-files/circle_on_cube.mesh.bz2 \
	 *		-refines1=4
	 */
	return;
    }

    if ((e0 != NULL || e1 != NULL) &&
	(e->mark > 0 || e->flag == CHECK_FAILED)) {
	if (e0 != NULL)
	    dof_interp0(e0);
	if (e1 != NULL)
	    dof_interp0(e1);
	return;
    }

    /* interpolate on e */
    p = dof_data_new;
    if (dof->count_vert > 0) {
	parent_data[ 0] = p + e->verts[0] * dof->count_vert;
	parent_data[ 1] = p + e->verts[1] * dof->count_vert;
	parent_data[ 2] = p + e->verts[2] * dof->count_vert;
	parent_data[ 3] = p + e->verts[3] * dof->count_vert;
	p += DofGetVertexDataCount(dof);
    }
    if (dof->count_edge > 0) {
	parent_data[ 4] = p + e->edges[0] * dof->count_edge;
	parent_data[ 5] = p + e->edges[1] * dof->count_edge;
	parent_data[ 6] = p + e->edges[2] * dof->count_edge;
	parent_data[ 7] = p + e->edges[3] * dof->count_edge;
	parent_data[ 8] = p + e->edges[4] * dof->count_edge;
	parent_data[ 9] = p + e->edges[5] * dof->count_edge;
	p += DofGetEdgeDataCount(dof);
    }
    if (dof->count_face > 0) {
	parent_data[10] = p + e->faces[0] * dof->count_face;
	parent_data[11] = p + e->faces[1] * dof->count_face;
	parent_data[12] = p + e->faces[2] * dof->count_face;
	parent_data[13] = p + e->faces[3] * dof->count_face;
	p += DofGetFaceDataCount(dof);
    }
    if (dof->count_elem > 0) {
	parent_data[14] = p + e->index * dof->count_elem;
    }
    dof_interp_recursive(e, parent_data);
}

static void
dof_interp(void)
{
    GRID *g = dof->g;
    INT i;

    dof_data_new = phgCalloc(DofGetDataCount(dof), sizeof(FLOAT));
    flags = phgCalloc((dof->count_vert > 0 ? g->nvert : 0) +
		      (dof->count_edge > 0 ? g->nedge : 0) +
		      (dof->count_face > 0 ? g->nface : 0), sizeof(BYTE));

    pass = 0;
    if (dof->userfunc != DofInterpolation) {
	for (i = 0; i < g->nroot; i++) {
	    dof_interp0(g->roots + i);
	}
	pass++;
    }

    for (i = 0; i < g->nroot; i++) {
	dof_interp0(g->roots + i);
    }

    phgFree(flags);
    phgFree(dof->data);

    dof->data = dof_data_new;
    dof->data_vert = dof->data;
    dof->data_edge = dof->data_vert + DofGetVertexDataCount(dof);
    dof->data_face = dof->data_edge + DofGetEdgeDataCount(dof);
    dof->data_elem = dof->data_face + DofGetFaceDataCount(dof);
}

/*----------------------- end DOF interpolation ------------------------*/

void
phgCoarsenMarkedElements(GRID *g)
/* coarsen all elements whose 'mark' member is negative */
{
    INT i;

    FunctionEntry_head;
    ParallelFunction(phgCoarsenMarkedElements, g != NULL && g->nprocs > 1);
    FunctionEntry_tail;

    if (g == NULL || g->nprocs <= phgRank)
	Return;

    if (g->nprocs > 1) {
	phgPrintf("\n"
"************************************************************************\n"
"* Parallel coarsening unavailable in this distribution, mesh unchanged *\n"
"************************************************************************\n"
	"\n");
	Return;
    }

    if (g->rank == 0)
	phgInfo(2, "before coarsening: nleaf=%d, nleaf_global=%d\n", 
		g->nleaf, g->nleaf_global);
    else
	phgDebug((3, "before coarsening: nleaf=%d, nleaf_global=%d\n", 
		g->nleaf, g->nleaf_global));

    g_ = g;

    /* set e->mark, reset e->flag, and set global indices */
    for (i = 0; i < g->nroot; i++)
	set_mark(g, g->roots + i);

    /* recursively set CHECK_FAILED mark on uncoarsenable patches */
    for (i = 0; i < g->nroot; i++)
	check_patch(g, g->roots + i);

    /* clear QUAD and DOF cache */
    phgDofClearCache(g,NULL,NULL, FALSE);
    dof_interp_flag = FALSE;
    if (g->dof != NULL) {
	DOF **pdof;
	/* note: loop in reverse order to keep g->geom valid on leaf elements */
	for (pdof = g->dof; *pdof != NULL; pdof++) ;
	for (; pdof-- > g->dof; ) {
	    dof = *pdof;
	    if (dof->data == NULL || SpecialDofType(dof->type) ||
		dof->type->InterpF2C == NULL || dof->userfunc == DofNoAction)
		continue;
	    dof_interp_flag = TRUE;
	    dof_interp();
	}
    }

    g->ntree = g->nleaf = 0;
    for (i = 0; i < g->nroot; i++) {
	do_coarsen(g, g->roots + i, &g->ntree, &g->nleaf);
    }

    if (g->nprocs == 1) {
	g->nelem_global = g->nleaf;
    }

    pack_indices(g, VERTEX);
    if (g->edge_flag)
	pack_indices(g, EDGE);
    if (g->face_flag)
	pack_indices(g, FACE);
    if (g->elem_flag)
	pack_indices(g, ELEMENT);

    /* adjust DOF data buffers and update pointers */
    if (g->dof != NULL) {
	DOF **pdof;
	for (pdof = g->dof; (dof = *pdof) != NULL; pdof++) {
	    if (dof->data == NULL || SpecialDofType(dof->type))
		continue;
	    if (dof->type->InterpF2C == NULL || dof->userfunc == DofNoAction) {
		phgFree(dof->data);
		dof->data = phgCalloc(DofGetDataCount(dof), sizeof(*dof->data));
	    }
	    else {
		dof->data = phgRealloc(dof->data, DofGetDataCount(dof)
							* sizeof(*dof->data));
	    }
	    dof->data_vert = dof->data;
	    dof->data_edge = dof->data_vert + DofGetVertexDataCount(dof);
	    dof->data_face = dof->data_edge + DofGetEdgeDataCount(dof);
	    dof->data_elem = dof->data_face + DofGetFaceDataCount(dof);
	}
    }

    update_neighbours(g);

    phgUpdateBoundaryTypes(g);

    /* TODO: if the grid is distributed, the DOFs may be inconsistent on
     * different submeshes, which may need to be corrected */

    if (g->rank == 0)
	phgInfo(2, "after coarsening:  nleaf=%d, nleaf_global=%d\n", 
		g->nleaf, g->nleaf_global);
    else
	phgDebug((2, "after coarsening:  nleaf=%d, nleaf_global=%d\n", 
		g->nleaf, g->nleaf_global));

    g->serial_no++;

    PrintTime(2);

    Return;
}
