/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2008 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: hierarchical-basis.c,v 1.67 2008/03/07 02:32:36 zlb Exp $
 *
 * This file implements general p-th order H1 conforming and Hcurl conforming
 * hierarchical basis functions */

#include "phg.h"

#include <stdlib.h>
#include <string.h>
#include <strings.h>	/* for bzero */
#include <math.h>

#define ORDER_MAX	15	/* maxi. polynomial order allowed */

/* Evaluating Legendre polynomials and derivatives using the recurrence
 * relation (P_l being the Legendre polynomial of order l):
 *
 *	P_0 = 1,
 *	P_1 = x,
 *	(l+1) * P_{l+1} - (2l+1) * x * P_l + l * P_{l-1} = 0,
 *	l = 1, 2, ...
 */

static void
legendre(int p, FLOAT x, FLOAT *values, FLOAT *values_d)
/* evaluates p-th order Legendre polynomial and its derivative:
 *	 values[k] = L_k(x),
 * and if values_d != NULL:
 *	values_d[k] = L'_k(x)
 */
{
    int i;
    FLOAT L0, L1, L2, DL0, DL1, DL2;

    if (p == 0) {
	*values = 1.0;
	if (values_d != NULL)
	    *values_d = 0.0;
	return;
    }
    else if (p == 1) {
	*(values++) = 1.0;
	*values = x;
	if (values_d != NULL) {
	    *(values_d++) = 0.0;
	    *values_d = 1.0;
	}
	return;
    }

    if (values_d == NULL) {
	*(values++) = L0 = 1.0;
	*(values++) = L1 = x;
	for (i = 1; i < p; i++) {
	    L2 = ((i + i + 1) * x * L1 - i * L0) / (FLOAT)(i + 1);
	    L0 = L1;
	    L1 = *(values++) = L2;
	}
    }
    else {
	*(values++) = L0 = 1.0;
	*(values++) = L1 = x;
	*(values_d++) = DL0 = 0.0;
	*(values_d++) = DL1 = 1.0;
	for (i = 1; i < p; i++) {
	    DL2 = ((i + i + 1) * (L1 + x * DL1) - i * DL0) / (FLOAT)(i + 1);
	    DL0 = DL1;
	    DL1 = *(values_d++) = DL2;
	    L2 = ((i + i + 1) * x * L1 - i * L0) / (FLOAT)(i + 1);
	    L0 = L1;
	    L1 = *(values++) = L2;
	}
    }

    return;
}

#if 0
/* testing evaluation of Legendre polynomials and derivatives */
static FLOAT
L10(FLOAT x)
{
    return -63 / (FLOAT)256.
	   +(3465. * Pow(x, 2)) / 256.
	   -(15015. * Pow(x, 4)) / 128.
	   +(45045. * Pow(x, 6)) / 128.
	   -(109395. * Pow(x, 8)) / 256.
	   +(46189. * Pow(x, 10)) / 256.;
}

static FLOAT
DL10(FLOAT x)
{
    return +(3465. * x) / 128.
	   -(15015. * Pow(x, 3)) / 32. 
	   +(135135. * Pow(x, 5)) / 64.
	   -(109395. * Pow(x, 7)) / 32.
	   +(230945. * Pow(x, 9)) / 128.;
}

int
main(int argc, char **argv[])
{
    int i, n = 10000;
    FLOAT x, d = 2.0 / n, L[11], DL[11];

    for (i = 0; i <= n; i++) {
	x = -1. + i * d;
	legendre(10, x, L, DL);
	/* Plot 1:2 and 1:3 for function/derivative, 1:4 and 1:5 for errors */
	printf("%lg %lg %lg %lg %lg\n",
		(double)x, (double)L[10], (double)DL[10],
		(double)(L[10] - L10(x)), (double)(DL[10] - DL10(x)));
    }

    return 0;
}
#endif	/* 0 */

/*-------------------------- Supporting functions --------------------------*/

/* buffers for caching evaluated Legendre polynomials. */

/* TODO: make use of symmetry of the basis functions and evaluation
 *	 points (mostly come from quadrature rules) to save more evaluations */

#define CACHE_LP 1	/* !0 ==> cache evaluated Legendre polynomials */

static FLOAT (*L0)[ORDER_MAX - 1] = NULL, (*DL0)[ORDER_MAX - 1] = NULL;

#if CACHE_LP
static BYTE *flags = NULL;
static BOOLEAN reset_flags = TRUE;

static inline void
get_L(int order, int v0, int v1, const FLOAT lambda[], FLOAT **L, FLOAT **DL)
/* evaluate and cache Legendre polynomials on a given edge */
{
    int no = GetEdgeNo(v0, v1);

    assert(order <= ORDER_MAX - 1);

    if (flags[no] == 0) {
	flags[no] = 1;
	legendre(order, lambda[v1] - lambda[v0],
		 L0[no], DL == NULL ? NULL : DL0[no]);
    }

    *L = L0[no];
    if (DL != NULL)
	*DL = DL0[no];
}
#endif	/* CACHE_LP */

/*-------------------------- H1 conforming bases --------------------------*/

static const FLOAT *
HB_bas(DOF *dof, SIMPLEX *e, int no0, int no1, const FLOAT *lambda)
{
    static FLOAT *buffers[ORDER_MAX + 1];
    static BOOLEAN initialized = FALSE;
    FLOAT *L1, *L2, *L3;

    DOF_TYPE *t = dof->type;
    int order = t->order;
    int nbas = t->nbas;
    FLOAT *bas, *p, *p0, beta;
    int i, j, k, i0, j0, v0, v1, v2, v3, m, n;

    if (!initialized) {
	initialized = TRUE;
	bzero(buffers, sizeof(buffers));
    }

    assert(order >= 1 && order <= ORDER_MAX);
    if (no1 <= 0)
	no1 = nbas;

    if ((bas = buffers[order]) == NULL) {
	buffers[order] = phgAlloc(nbas * sizeof(*bas));
	FreeAtExit(buffers[order]);
	bas = buffers[order];
    }

    if (order == 0) {
	*bas = 1.0;
	return bas;
    }

    p = bas;
    p0 = p + (no1 - no0);

    /*------------------------- vertex functions --------------------------*/
    for (i = no0; i < NVert && i < no1; i++)
	*(p++) = lambda[i];

    if (i >= no1)
	return bas;

#if CACHE_LP
    if (L0 == NULL) {
	L0 = phgAlloc(NEdge * sizeof(*L0));
	DL0 = phgAlloc(NEdge * sizeof(*DL0));
	flags = phgAlloc(NEdge * sizeof(*flags));
	FreeAtExit(L0);
	FreeAtExit(DL0);
	FreeAtExit(flags);
    }
    if (reset_flags)
	memset(flags, 0, NEdge * sizeof(*flags));
#else	/* CACHE_LP */
    if (L0 == NULL) {
	L0 = phgAlloc(3 * sizeof(*L0));
	DL0 = phgAlloc(3 * sizeof(*DL0));
	FreeAtExit(L0);
	FreeAtExit(DL0);
    }
#endif	/* CACHE_LP */

    /*------------------------- edge functions --------------------------*/
    if ((no0 -= NVert) < 0)
	no0 = 0;
    n = t->np_edge;
    i0 = no0 / n;
    j0 = no0 % n;
    for (i = i0; i < NEdge && p < p0; i++) {
	GetEdgeVertices(e, i, v0, v1);
	beta = lambda[v0] * lambda[v1];
#if CACHE_LP
	get_L(order - 2, v0, v1, lambda, &L1, NULL);
#else	/* CACHE_LP */
	legendre(order - 2, lambda[v1] - lambda[v0], L1 = L0[0], NULL);
#endif	/* CACHE_LP */
	for (j = j0; j < n && p < p0; j++)
	    *(p++) = beta * L1[j];
	j0 = 0;
    }

    if (p >= p0)
	return bas;

    /*------------------------- face functions --------------------------*/
    if ((no0 -= NEdge * n) < 0)
	no0 = 0;
    n = t->np_face;
    i0 = no0 / n;
    j0 = no0 % n;
    for (i = i0; i < NFace && p < p0; i++) {
	GetFaceVertices(e, i, v0, v1, v2);
	beta = lambda[v0] * lambda[v1] * lambda[v2];
#if CACHE_LP
	get_L(order - 3, v0, v1, lambda, &L1, NULL);
	get_L(order - 3, v0, v2, lambda, &L2, NULL);
#else	/* CACHE_LP */
	legendre(order - 3, lambda[v1] - lambda[v0], L1 = L0[0], NULL);
	legendre(order - 3, lambda[v2] - lambda[v0], L2 = L0[1], NULL);
#endif	/* CACHE_LP */
	for (m = 0; m <= order - 3 && p < p0; m++) {
	    if (j0 >= m + 1) {
		j0 -= m + 1;
		continue;
	    }
	    for (j = j0; j <= m && p < p0; j++)
		*(p++) = beta * L1[j] * L2[m - j];
	    j0 = 0;
	}
    }

    if (p >= p0)
	return bas;

    /*------------------------- element functions --------------------------*/
    if ((no0 -= NFace * n) < 0)
	no0 = 0;
    n = t->np_elem;
    assert(no0 < n);
    j0 = no0;
    GetElementVertices(e, 0, v0, v1, v2, v3);
    beta = lambda[v0] * lambda[v1] * lambda[v2] * lambda[v3];
#if CACHE_LP
    get_L(order - 4, v0, v1, lambda, &L1, NULL);
    get_L(order - 4, v0, v2, lambda, &L2, NULL);
    get_L(order - 4, v0, v3, lambda, &L3, NULL);
#else	/* CACHE_LP */
    legendre(order - 4, lambda[v1] - lambda[v0], L1 = L0[0], NULL);
    legendre(order - 4, lambda[v2] - lambda[v0], L2 = L0[1], NULL);
    legendre(order - 4, lambda[v3] - lambda[v0], L3 = L0[2], NULL);
#endif	/* CACHE_LP */
    for (m = 0; m <= order - 4 && p < p0; m++) {
	if (j0 >= (m + 1) * (m + 2) / 2) {
	    j0 -= (m + 1) * (m + 2) / 2;
	    continue;
	}
	for (k = 0; k <= m && p < p0; k++) {
	    if (j0 >= m - k + 1) {
		j0 -= m - k + 1;
		continue;
	    }
	    for (j = j0; j <= m - k && p < p0; j++)
		*(p++) = beta * L1[j] * L2[k] * L3[m - j - k];
	    j0 = 0;
	}
    }

    if (p != p0)
	phgError(1, "HB_bas: p=%d, p0=%d\n", (int)(p - bas), (int)(p0 - bas));

    return bas;
}

static const FLOAT *
HB_grad(DOF *dof, SIMPLEX *e, int no0, int no1, const FLOAT *lambda)
{
    static FLOAT *buffers[ORDER_MAX + 1];
    static BOOLEAN initialized = FALSE;
    FLOAT *L1, *L2, *L3, *DL1, *DL2, *DL3;

    DOF_TYPE *t = dof->type;
    int order = t->order;
    int ngrad = (Dim + 1) * t->nbas;
    FLOAT *grad, *p, *p0, beta, beta0, beta1, beta2, beta3, a, b, c, d;
    int i, j, k, ii, i0, j0, v0, v1, v2, v3, m, n;

    if (!initialized) {
	initialized = TRUE;
	bzero(buffers, sizeof(buffers));
    }

    assert(order >= 1 && order <= ORDER_MAX);
    if (no1 <= 0)
	no1 = t->nbas;

    if ((grad = buffers[order]) == NULL) {
	buffers[order] = phgAlloc(ngrad * sizeof(*grad));
	FreeAtExit(buffers[order]);
	grad = buffers[order];
    }

    if (order == 0) {
	grad[0] = grad[1] = grad[2] = grad[3] = 0.0;
	return grad;
    }

    p = grad;
    p0 = p + (no1 - no0) * (Dim + 1);

    /*------------------------- vertex functions --------------------------*/
    for (i = no0; i < NVert && i < no1; i++)
	for (j = 0; j < Dim + 1; j++)
	    *(p++) = (i == j ? 1.0 : 0.0);

    if (i >= no1)
	return grad;

#if CACHE_LP
    if (L0 == NULL) {
	L0 = phgAlloc(NEdge * sizeof(*L0));
	DL0 = phgAlloc(NEdge * sizeof(*DL0));
	flags = phgAlloc(NEdge * sizeof(*flags));
	FreeAtExit(L0);
	FreeAtExit(DL0);
	FreeAtExit(flags);
    }
    if (reset_flags)
	memset(flags, 0, NEdge * sizeof(*flags));
#else	/* CACHE_LP */
    if (L0 == NULL) {
	L0 = phgAlloc(3 * sizeof(*L0));
	DL0 = phgAlloc(3 * sizeof(*DL0));
	FreeAtExit(L0);
	FreeAtExit(DL0);
    }
#endif	/* CACHE_LP */

    /*------------------------- edge functions --------------------------*/
    if ((no0 -= NVert) < 0)
	no0 = 0;
    n = t->np_edge;
    i0 = no0 / n;
    j0 = no0 % n;
    for (i = i0; i < NEdge && p < p0; i++) {
	GetEdgeVertices(e, i, v0, v1);
	beta = lambda[v0] * lambda[v1];
#if CACHE_LP
	get_L(order - 2, v0, v1, lambda, &L1, &DL1);
#else	/* CACHE_LP */
	legendre(order - 2, lambda[v1] - lambda[v0], L1 = L0[0], DL1 = DL0[0]);
#endif	/* CACHE_LP */
	for (j = j0; j < n && p < p0; j++) {
	    a = beta * DL1[j];
	    for (ii = 0; ii < Dim + 1; ii++) {
		if (ii == v0)
		    *(p++) = lambda[v1] * L1[j] - a;
		else if (ii == v1)
		    *(p++) = lambda[v0] * L1[j] + a;
		else
		    *(p++) = 0.0;
	    }
	}
	j0 = 0;
    }

    if (p >= p0)
	return grad;

    /*------------------------- face functions --------------------------*/
    if ((no0 -= NEdge * n) < 0)
	no0 = 0;
    n = t->np_face;
    i0 = no0 / n;
    j0 = no0 % n;
    for (i = i0; i < NFace && p < p0; i++) {
	GetFaceVertices(e, i, v0, v1, v2);
	beta0 = lambda[v1] * lambda[v2];
	beta1 = lambda[v0] * lambda[v2];
	beta2 = lambda[v0] * lambda[v1];
	beta = lambda[v0] * beta0;
#if CACHE_LP
	get_L(order - 3, v0, v1, lambda, &L1, &DL1);
	get_L(order - 3, v0, v2, lambda, &L2, &DL2);
#else	/* CACHE_LP */
	legendre(order - 3, lambda[v1] - lambda[v0], L1 = L0[0], DL1 = DL0[0]);
	legendre(order - 3, lambda[v2] - lambda[v0], L2 = L0[1], DL2 = DL0[1]);
#endif	/* CACHE_LP */
	for (m = 0; m <= order - 3 && p < p0; m++) {
	    if (j0 >= m + 1) {
		j0 -= m + 1;
		continue;
	    }
	    for (j = j0; j <= m && p < p0; j++) {
		a = L1[j] * L2[m - j];
		b = beta * DL1[j] * L2 [m - j];
		c = beta * L1 [j] * DL2[m - j];
		for (ii = 0; ii < Dim + 1; ii++) {
		    if (ii == v0)
			*(p++) = beta0 * a - (b + c);
		    else if (ii == v1)
			*(p++) = beta1 * a + b;
		    else if (ii == v2)
			*(p++) = beta2 * a + c;
		    else
			*(p++) = 0.;
		}
	    }
	    j0 = 0;
	}
    }

    if (p >= p0)
	return grad;

    /*------------------------- element functions --------------------------*/
    if ((no0 -= NFace * n) < 0)
	no0 = 0;
    n = t->np_elem;
    assert(no0 < n);
    j0 = no0;
    GetElementVertices(e, 0, v0, v1, v2, v3);
    beta0 = lambda[v1] * lambda[v2] * lambda[v3];
    beta1 = lambda[v0] * lambda[v2] * lambda[v3];
    beta2 = lambda[v0] * lambda[v1] * lambda[v3];
    beta3 = lambda[v0] * lambda[v1] * lambda[v2];
    beta = lambda[v0] * beta0;
#if CACHE_LP
    get_L(order - 4, v0, v1, lambda, &L1, &DL1);
    get_L(order - 4, v0, v2, lambda, &L2, &DL2);
    get_L(order - 4, v0, v3, lambda, &L3, &DL3);
#else	/* CACHE_LP */
    legendre(order - 4, lambda[v1] - lambda[v0], L1 = L0[0], DL1 = DL0[0]);
    legendre(order - 4, lambda[v2] - lambda[v0], L2 = L0[1], DL2 = DL0[1]);
    legendre(order - 4, lambda[v3] - lambda[v0], L3 = L0[2], DL3 = DL0[2]);
#endif	/* CACHE_LP */
    for (m = 0; m <= order - 4 && p < p0; m++) {
	if (j0 >= (m + 1) * (m + 2) / 2) {
	    j0 -= (m + 1) * (m + 2) / 2;
	    continue;
	}
	for (k = 0; k <= m && p < p0; k++) {
	    if (j0 >= m - k + 1) {
		j0 -= m - k + 1;
		continue;
	    }
	    for (j = j0; j <= m - k && p < p0; j++) {
		a = L1[j] * L2[k] * L3[m - j - k];
		b = beta * DL1[j] * L2 [k] * L3 [m - j - k];
		c = beta * L1 [j] * DL2[k] * L3 [m - j - k];
		d = beta * L1 [j] * L2 [k] * DL3[m - j - k];
		p[v0] = beta0 * a - (b + c + d);
		p[v1] = beta1 * a + b;
		p[v2] = beta2 * a + c;
		p[v3] = beta3 * a + d;
		p += Dim + 1;
	    }
	    j0 = 0;
	}
    }

    assert(p == p0);

    return grad;
}

static DOF *dof0;

static void
HB_init0(DOF *dof, SIMPLEX *e, GTYPE type, int index,
	DOF_USER_FUNC userfunc, DOF_USER_FUNC_LAMBDA userfunc_lambda,
	const FLOAT *funcvalues, FLOAT *dofvalues, FLOAT **pdofvalues)
{
    DOF_TYPE *t = dof->type;
    int order = t->order;
    FLOAT a, b, w, x, y, z, *mat, lambda[Dim + 1] = {0., 0., 0., 0.};
    /* pointers to (known) DOF values at lower dimensional entities */
    const FLOAT *f0, *f1, *f2, *f3,		/* vertices */
		*g0, *g1, *g2, *g3, *g4, *g5,	/* edges */
		*h0, *h1, *h2, *h3;		/* faces */
    FLOAT *f = NULL, *qp, *qw;
    const FLOAT *bas, *p;
    int i, j, k, n, m, no0, no1, ii, np2, np3, *pvt = NULL;
    int v0, v1, v2, v3, i0, i1, i2, i3, i4, i5;
    QUAD *quad;

    /* cache for factorized matrices */
    static FLOAT *LU_cache[3 * (ORDER_MAX + 1)];
    static int *PV_cache[3 * (ORDER_MAX + 1)];
    static BOOLEAN initialized = FALSE;

    if (!initialized) {
	initialized = TRUE;
	bzero(LU_cache, sizeof(LU_cache));
	bzero(PV_cache, sizeof(PV_cache));
    }

    assert(order >= 1 && order <= ORDER_MAX);
    assert(userfunc != NULL || userfunc_lambda != NULL);

    switch (type) {
	case VERTEX:
	    /* assign function values to DOF */
	    assert(index >= 0 && index < NVert && t->np_vert > 0);
	    lambda[index] = 1.0;
	    if (userfunc_lambda != NULL) {
		userfunc_lambda(dof0, e, 0, lambda, dofvalues);
	    }
	    else {
		phgGeomLambda2XYZ(dof->g, e, lambda, &x, &y, &z);
		userfunc(x, y, z, dofvalues);
	    }
	    break;
	case EDGE:
	    /* compute DOF by L2-projection on the edge */
	    assert(index >= 0 && index < NEdge && t->np_edge > 0);
	    GetEdgeVertices(e, index, v0, v1)
	    n = t->np_edge;
	    m = dof->dim;
	    f = phgAlloc(m * sizeof(*f));
	    if (pdofvalues == NULL) {
		f0 = DofVertexData(dof, e->verts[v0]);
		f1 = DofVertexData(dof, e->verts[v1]);
	    }
	    else {
		f0 = pdofvalues[v0];
		f1 = pdofvalues[v1];
	    }
	    /* compute the DOF by solving dof->dim systems of linear equations
	     * of size n, dof->data or pdofvalues[] should point to valid
	     * DOF values on the two vertices */
	    if (LU_cache[3 * order] == NULL) {
		LU_cache[3 * order] = phgCalloc(n * n, sizeof(*mat));
		FreeAtExit(LU_cache[3 * order]);
		PV_cache[3 * order] = phgAlloc(n * sizeof(*pvt));
		FreeAtExit(PV_cache[3 * order]);
		mat = LU_cache[3 * order];
		pvt = PV_cache[3 * order];
	    }
	    else {
		mat = NULL;
	    }
	    bzero(dofvalues, m * n * sizeof(*dofvalues));
	    quad = phgQuadGetQuad1D(2 * order);
	    qp = quad->points;
	    qw = quad->weights;
	    no0 = NVert + index * n;
	    no1 = no0 + n;
#define A(i,j) mat[(i) * n + (j)]
#define B(i,j) dofvalues[(i) * m + (j)]
	    for (k = 0; k < quad->npoints; k++) {
		lambda[v0] = *(qp++);
		lambda[v1] = *(qp++);
		w = *(qw++);
		if (userfunc_lambda != NULL) {
		    userfunc_lambda(dof0, e, 0, lambda, f);
		}
		else {
		    phgGeomLambda2XYZ(dof->g, e, lambda, &x, &y, &z);
		    userfunc(x, y, z, f);
		}
		/* Note: must call BasFuncs after user_func since the latter
		 * might call BasFuncs, such as with phgDofInterpC2FGeneric */
		bas = t->BasFuncs(dof, e, no0, no1, lambda);
		for (i = 0; i < n; i++) {
		    b = bas[i] * w;
		    if (mat != NULL) {
			for (j = 0; j <= i; j++) {
			    A(i,j) += (a = b * bas[j]);
			    if (j < i)
				A(j,i) += a;
			}
		    }
		    for (j = 0; j < m; j++) {
			a = f[j] - (lambda[v0] * f0[j] + lambda[v1] * f1[j]);
			B(i,j) += b * a;
		    }
		}
	    }
#if 0
if (mat != NULL) {
printf("---------------------------------------------\n");
for (i = 0; i < n; i++) {
for (j = 0; j < n; j++) printf(" %0.4lf",
		(double)(Fabs(A(i,j))<1e-10 ? 0. : A(i,j)));
printf(" = %lg\n", (double)B(i,0));
}
printf("---------------------------------------------\n");
}
#endif
#undef A
#undef B
	    /* solve the equation A x = B */
	    if (mat != NULL) {
		if (!phgSolverDenseLU(n, (void *)mat, pvt))
		    phgError(1, "%s:%d: singular system!\n",__FILE__,__LINE__);
	    }
	    else {
		mat = LU_cache[3 * order];
		pvt = PV_cache[3 * order];
	    }
	    phgSolverDenseSV(n, (void *)mat, pvt, m, (void *)dofvalues);
	    break;
	case FACE:
	    /* compute DOF by L2-projection on the face */
	    assert(index >= 0 && index < NFace && t->np_face > 0);
	    np2 = t->np_edge;
	    GetFaceVertices(e, index, v0, v1, v2)
	    n = t->np_face;
	    m = dof->dim;
	    f = phgAlloc(m * sizeof(*f));
	    /* the next 2 lines make gcc happy */
	    f0 = f1 = f2 = g0 = g1 = g2 = NULL;
	    i0 = i1 = i2 = 0;
	    if (pdofvalues == NULL) {
		f0 = DofVertexData(dof, e->verts[v0]);
		f1 = DofVertexData(dof, e->verts[v1]);
		f2 = DofVertexData(dof, e->verts[v2]);
		g0 = DofEdgeData(dof, e->edges[i0 = GetEdgeNo(v0, v1)]);
		g1 = DofEdgeData(dof, e->edges[i1 = GetEdgeNo(v0, v2)]);
		g2 = DofEdgeData(dof, e->edges[i2 = GetEdgeNo(v1, v2)]);
	    }
	    else {
		f0 = pdofvalues[v0];
		f1 = pdofvalues[v1];
		f2 = pdofvalues[v2];
		g0 = pdofvalues[NVert + (i0 = GetEdgeNo(v0, v1))];
		g1 = pdofvalues[NVert + (i1 = GetEdgeNo(v0, v2))];
		g2 = pdofvalues[NVert + (i2 = GetEdgeNo(v1, v2))];
	    }
	    /* compute the DOF by solving dof->dim systems of linear equations
	     * of size n, dof->data or pdofvalues[] should point to valid
	     * DOF values on the two vertices */
	    if (LU_cache[3 * order + 1] == NULL) {
		LU_cache[3 * order + 1] = phgCalloc(n * n, sizeof(*mat));
		FreeAtExit(LU_cache[3 * order + 1]);
		PV_cache[3 * order + 1] = phgAlloc(n * sizeof(*pvt));
		FreeAtExit(PV_cache[3 * order + 1]);
		mat = LU_cache[3 * order + 1];
		pvt = PV_cache[3 * order + 1];
	    }
	    else {
		mat = NULL;
	    }
	    bzero(dofvalues, m * n * sizeof(*dofvalues));
	    quad = phgQuadGetQuad2D(2 * order);
	    qp = quad->points;
	    qw = quad->weights;
	    no0 = NVert + NEdge * np2 + index * n;
	    no1 = no0 + n;
#define A(i,j) mat[(i) * n + (j)]
#define B(i,j) dofvalues[(i) * m + (j)]
	    for (k = 0; k < quad->npoints; k++) {
		lambda[v0] = *(qp++);
		lambda[v1] = *(qp++);
		lambda[v2] = *(qp++);
		w = *(qw++);
		if (userfunc_lambda != NULL) {
		    userfunc_lambda(dof0, e, 0, lambda, f);
		}
		else {
		    phgGeomLambda2XYZ(dof->g, e, lambda, &x, &y, &z);
		    userfunc(x, y, z, f);
		}
		/* Note: must call BasFuncs after user_func since the latter
		 * might call BasFuncs, such as with phgDofInterpC2FGeneric */
		bas = t->BasFuncs(dof, e, NVert, no1, lambda);
		p = bas + no0 - NVert;
		for (i = 0; i < n; i++) {
		    b = p[i] * w;
		    if (mat != NULL) {
			for (j = 0; j <= i; j++) {
			    A(i,j) += (a = b * p[j]);
			    if (j < i)
				A(j,i) += a;
			}
		    }
		    for (j = 0; j < m; j++) {
			a = f[j] - (lambda[v0] * f0[j] + lambda[v1] * f1[j] +
				    lambda[v2] * f2[j]);
			for (ii = 0; ii < np2; ii++)
			    a -= bas[i0 * np2 + ii] * g0[ii * m + j] +
				 bas[i1 * np2 + ii] * g1[ii * m + j] +
				 bas[i2 * np2 + ii] * g2[ii * m + j];
			B(i,j) += b * a;
		    }
		}
	    }
#undef A
#undef B
	    /* solve the equation A x = B */
	    if (mat != NULL) {
		if (!phgSolverDenseLU(n, (void *)mat, pvt))
		    phgError(1, "%s:%d: singular system!\n",__FILE__,__LINE__);
	    }
	    else {
		mat = LU_cache[3 * order + 1];
		pvt = PV_cache[3 * order + 1];
	    }
	    phgSolverDenseSV(n, (void *)mat, pvt, m, (void *)dofvalues);
	    break;
	case ELEMENT:
	    /* compute DOF by L2-projection on the element */
	    assert(index == 0 && t->np_elem > 0);
	    np2 = t->np_edge;
	    np3 = t->np_face;
	    GetElementVertices(e, 0, v0, v1, v2, v3);
	    n = t->np_elem;
	    m = dof->dim;
	    f = phgAlloc(m * sizeof(*f));
	    /* the next 3 lines make gcc happy */
	    f0 = f1 = f2 = f3 = g0 = g1 = g2 = g3 = g4 = g5 = h0 = h1 = h2 = h3
		= NULL;
	    i0 = i1 = i2 = i3 = i4 = i5 = 0;
	    if (pdofvalues == NULL) {
		f0 = DofVertexData(dof, e->verts[v0]);
		f1 = DofVertexData(dof, e->verts[v1]);
		f2 = DofVertexData(dof, e->verts[v2]);
		f3 = DofVertexData(dof, e->verts[v3]);
		g0 = DofEdgeData(dof, e->edges[i0 = GetEdgeNo(v0, v1)]);
		g1 = DofEdgeData(dof, e->edges[i1 = GetEdgeNo(v0, v2)]);
		g2 = DofEdgeData(dof, e->edges[i2 = GetEdgeNo(v0, v3)]);
		g3 = DofEdgeData(dof, e->edges[i3 = GetEdgeNo(v1, v2)]);
		g4 = DofEdgeData(dof, e->edges[i4 = GetEdgeNo(v1, v3)]);
		g5 = DofEdgeData(dof, e->edges[i5 = GetEdgeNo(v2, v3)]);
		h0 = DofFaceData(dof, e->faces[v0]);
		h1 = DofFaceData(dof, e->faces[v1]);
		h2 = DofFaceData(dof, e->faces[v2]);
		h3 = DofFaceData(dof, e->faces[v3]);
	    }
	    else {
		f0 = pdofvalues[v0];
		f1 = pdofvalues[v1];
		f2 = pdofvalues[v2];
		f3 = pdofvalues[v3];
		g0 = pdofvalues[NVert + (i0 = GetEdgeNo(v0, v1))];
		g1 = pdofvalues[NVert + (i1 = GetEdgeNo(v0, v2))];
		g2 = pdofvalues[NVert + (i2 = GetEdgeNo(v0, v3))];
		g3 = pdofvalues[NVert + (i3 = GetEdgeNo(v1, v2))];
		g4 = pdofvalues[NVert + (i4 = GetEdgeNo(v1, v3))];
		g5 = pdofvalues[NVert + (i5 = GetEdgeNo(v2, v3))];
		h0 = pdofvalues[NVert + NEdge + v0];
		h1 = pdofvalues[NVert + NEdge + v1];
		h2 = pdofvalues[NVert + NEdge + v2];
		h3 = pdofvalues[NVert + NEdge + v3];
	    }
	    /* compute the DOF by solving dof->dim systems of linear equations
	     * of size n, dof->data or pdofvalues[] should point to valid
	     * DOF values on the two vertices */
	    if (LU_cache[3 * order + 2] == NULL) {
		LU_cache[3 * order + 2] = phgCalloc(n * n, sizeof(*mat));
		FreeAtExit(LU_cache[3 * order + 2]);
		PV_cache[3 * order + 2] = phgAlloc(n * sizeof(*pvt));
		FreeAtExit(PV_cache[3 * order + 2]);
		mat = LU_cache[3 * order + 2];
		pvt = PV_cache[3 * order + 2];
	    }
	    else {
		mat = NULL;
	    }
	    bzero(dofvalues, m * n * sizeof(*dofvalues));
	    quad = phgQuadGetQuad3D(2 * order);
	    qp = quad->points;
	    qw = quad->weights;
	    no0 = NVert + NEdge * np2 + NFace * np3;
	    no1 = no0 + n;
#define A(i,j) mat[(i) * n + (j)]
#define B(i,j) dofvalues[(i) * m + (j)]
	    for (k = 0; k < quad->npoints; k++) {
		lambda[v0] = *(qp++);
		lambda[v1] = *(qp++);
		lambda[v2] = *(qp++);
		lambda[v3] = *(qp++);
		w = *(qw++);
		if (userfunc_lambda != NULL) {
		    userfunc_lambda(dof0, e, 0, lambda, f);
		}
		else {
		    phgGeomLambda2XYZ(dof->g, e, lambda, &x, &y, &z);
		    userfunc(x, y, z, f);
		}
		/* Note: must call BasFuncs after user_func since the latter
		 * might call BasFuncs, such as with phgDofInterpC2FGeneric */
		bas = t->BasFuncs(dof, e, NVert, no1, lambda);
		p = bas + no0 - NVert;
		for (i = 0; i < n; i++) {
		    b = p[i] * w;
		    if (mat != NULL) {
			for (j = 0; j <= i; j++) {
			    A(i,j) += (a = b * p[j]);
			    if (j < i)
				A(j,i) += a;
			}
		    }
		    for (j = 0; j < m; j++) {
			a = f[j] - (lambda[v0] * f0[j] + lambda[v1] * f1[j] +
				    lambda[v2] * f2[j] + lambda[v3] * f3[j]);
			for (ii = 0; ii < np2; ii++)
			    a -= bas[i0 * np2 + ii] * g0[ii * m + j] +
				 bas[i1 * np2 + ii] * g1[ii * m + j] +
				 bas[i2 * np2 + ii] * g2[ii * m + j] +
				 bas[i3 * np2 + ii] * g3[ii * m + j] +
				 bas[i4 * np2 + ii] * g4[ii * m + j] +
				 bas[i5 * np2 + ii] * g5[ii * m + j];
			for (ii = 0; ii < np3; ii++)
			    a -= bas[NEdge*np2 + v0*np3 + ii] * h0[ii*m + j] +
				 bas[NEdge*np2 + v1*np3 + ii] * h1[ii*m + j] +
				 bas[NEdge*np2 + v2*np3 + ii] * h2[ii*m + j] +
				 bas[NEdge*np2 + v3*np3 + ii] * h3[ii*m + j];
			B(i,j) += b * a;
		    }
		}
	    }
#undef A
#undef B
	    /* solve the equation A x = B */
	    if (mat != NULL) {
		if (!phgSolverDenseLU(n, (void *)mat, pvt))
		    phgError(1, "%s:%d: singular system!\n",__FILE__,__LINE__);
	    }
	    else {
		mat = LU_cache[3 * order + 2];
		pvt = PV_cache[3 * order + 2];
	    }
	    phgSolverDenseSV(n, (void *)mat, pvt, m, (void *)dofvalues);
	    break;
    }

    phgFree(f);
    return;
}

static void
HB_init(DOF *dof, SIMPLEX *e, GTYPE type, int index,
	DOF_USER_FUNC userfunc, DOF_USER_FUNC_LAMBDA userfunc_lambda,
	const FLOAT *funcvalues, FLOAT *dofvalues, FLOAT **pdofvalues)
{
    DOF_TYPE *t = dof->type;
    FLOAT *pv[15];
    int i, m = dof->dim;

    if (t->np_vert == 1) {
	dof0 = dof;
	HB_init0(dof, e, type, index, userfunc, userfunc_lambda,
		 funcvalues, dofvalues, pdofvalues);
	return;
    }

    /* map DG basis functions to HB basis functions */
    assert(type == ELEMENT);
    dof0 = phgAlloc(sizeof(*dof0));
    memcpy(dof0, dof, sizeof(*dof));
    dof->type = DOF_HBn[t->order];

    for (i = 0; i < NVert; i++) {
	HB_init0(dof, e, VERTEX, i, userfunc, userfunc_lambda,
		 funcvalues, dofvalues, NULL);
	pv[i] = dofvalues;
	dofvalues += m;
    }

    for (i = 0; i < NEdge; i++) {
	HB_init0(dof, e, EDGE, i, userfunc, userfunc_lambda,
		 funcvalues, dofvalues, pv);
	pv[NVert + i] = dofvalues;
	dofvalues += m * dof->type->np_edge;
    }

    for (i = 0; i < NFace; i++) {
	HB_init0(dof, e, FACE, i, userfunc, userfunc_lambda,
		 funcvalues, dofvalues, pv);
	pv[NVert + NEdge + i] = dofvalues;
	dofvalues += m * dof->type->np_face;
    }

    HB_init0(dof, e, ELEMENT, 0, userfunc, userfunc_lambda,
	     funcvalues, dofvalues, pv);

    dof->type = t;
    phgFree(dof0);
}

/*-------------------------- DOF types HBn --------------------------*/

#define NBasEdge(p)	(p > 1 ? (p - 1) : 0)
#define NBasFace(p)	(p > 2 ? (p - 1) * (p - 2) / 2 : 0)
#define NBasElem(p)	(p > 3 ? (p - 1) * (p - 2) * (p - 3) / 6 : 0)

#define	D1(a)	a
#define	D2(a)	a, a
#define	D3(a)	a, D2(a)
#define	D4(a)	D2(a), D2(a)
#define	D5(a)	D2(a), D3(a)
#define	D6(a)	D3(a), D3(a)
#define	D7(a)	D3(a), D4(a)
#define	D8(a)	D4(a), D4(a)
#define	D9(a)	D4(a), D5(a)
#define	D10(a)	D5(a), D5(a)
#define	D11(a)	D5(a), D6(a)
#define	D12(a)	D6(a), D6(a)
#define	D13(a)	D6(a), D7(a)
#define	D14(a)	D7(a), D7(a)
#define	D15(a)	D7(a), D8(a)
#define D21(a)  D15(a), D6(a)
#define D28(a)  D21(a), D7(a)
#define D36(a)  D28(a), D8(a)
#define D45(a)  D36(a), D9(a)
#define D55(a)  D45(a), D10(a)
#define D66(a)  D55(a), D11(a)
#define D78(a)  D66(a), D12(a)

#define NBasEdge(p)	(p > 1 ? (p - 1) : 0)
#define NBasFace(p)	(p > 2 ? (p - 1) * (p - 2) / 2 : 0)
#define NBasElem(p)	(p > 3 ? (p - 1) * (p - 2) * (p - 3) / 6 : 0)

static HP_INFO hp_info_HB = {DOF_HBn, 1, 15};

#define DofType(order, type, name, orders, grad, invariant)		\
    DOF_TYPE type = {DofCache, name, NULL, orders,			\
			grad, NULL, &hp_info_HB,			\
			phgDofInterpC2FGeneric,				\
			phgDofInterpF2CGeneric,				\
			HB_init, HB_bas, HB_grad,			\
			invariant, FALSE, -1,				\
			NVert + NEdge * NBasEdge(order)	+		\
			    NFace * NBasFace(order) + NBasElem(order),	\
			order,		/* order */			\
			0,		/* continuity */		\
			1,		/* dim */			\
			1, NBasEdge(order), NBasFace(order), NBasElem(order)};

static BYTE HB1_orders[] = {1};
DofType(1, DOF_HB1_, "HB1", HB1_orders, DOF_DG0, TRUE)

static BYTE HB2_orders[] = {
    1,		/* vertex */
    2		/* edge */
};
DofType(2, DOF_HB2_, "HB2", HB2_orders, DOF_DG1, TRUE)

static BYTE HB3_orders[] = {/* vertex: */ 1, /* edge: */ 2, 3, /* face: */ 3};
DofType(3, DOF_HB3_, "HB3", HB3_orders, DOF_DG2, FALSE)

static BYTE HB4_orders[] = {
    1,		/* vertex */
    2, 3, 4,	/* edge */
    3, D2(4),	/* face */
    4		/* element */
};
DofType(4, DOF_HB4_, "HB4", HB4_orders, DOF_DG3, FALSE)

static BYTE HB5_orders[] = {
    1,			/* vertex */
    2, 3, 4, 5,		/* edge */
    3, D2(4), D3(5),	/* face */
    4, D3(5)		/* element (Note: Dn(k), n = (k-2)*(k-3)/2) */
};
DofType(5, DOF_HB5_, "HB5", HB5_orders, DOF_DG4, FALSE)

static BYTE HB6_orders[] = {
    1,				/* vertex */
    2, 3, 4, 5, 6,		/* edge */
    3, D2(4), D3(5), D4(6),	/* face */
    4, D3(5), D6(6) 		/* element */
};
DofType(6, DOF_HB6_, "HB6", HB6_orders, DOF_DG5, FALSE)

static BYTE HB7_orders[] = {
    1,					/* vertex */
    2, 3, 4, 5, 6, 7,			/* edge */
    3, D2(4), D3(5), D4(6), D5(7),	/* face */
    4, D3(5), D6(6), D10(7)		/* element */
};
DofType(7, DOF_HB7_, "HB7", HB7_orders, DOF_DG6, FALSE)

static BYTE HB8_orders[] = {
    1,						/* vertex */
    2, 3, 4, 5, 6, 7, 8,			/* edge */
    3, D2(4), D3(5), D4(6), D5(7), D8(8),	/* face */
    4, D3(5), D6(6), D10(7), D15(8)		/* element */
};
DofType(8, DOF_HB8_, "HB8", HB8_orders, DOF_DG7, FALSE)

static BYTE HB9_orders[] = {
    1,							/* vertex */
    2, 3, 4, 5, 6, 7, 8, 9,				/* edge */
    3, D2(4), D3(5), D4(6), D5(7), D8(8), D9(9),	/* face */
    4, D3(5), D6(6), D10(7), D15(8), D21(9)		/* element */
};
DofType(9, DOF_HB9_, "HB9", HB9_orders, DOF_DG8, FALSE)

static BYTE HB10_orders[] = {
    1,								/* vertex */
    2, 3, 4, 5, 6, 7, 8, 9, 10,					/* edge */
    3, D2(4), D3(5), D4(6), D5(7), D8(8), D9(9), D10(10),	/* face */
    4, D3(5), D6(6), D10(7), D15(8), D21(9), D28(10)		/* element */
};
DofType(10, DOF_HB10_, "HB10", HB10_orders, DOF_DG9, FALSE)

static BYTE HB11_orders[] = {
    1,								/* vertex */
    2, 3, 4, 5, 6, 7, 8, 9, 10, 11,				/* edge */
    3, D2(4), D3(5), D4(6), D5(7), D8(8), D9(9), D10(10), D11(11),
								/* face */
    4, D3(5), D6(6), D10(7), D15(8), D21(9), D28(10), D36(11)	/* element */
};
DofType(11, DOF_HB11_, "HB11", HB11_orders, DOF_DG10, FALSE)

static BYTE HB12_orders[] = {
    1,								/* vertex */
    2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12,				/* edge */
    3, D2(4), D3(5), D4(6), D5(7), D8(8), D9(9), D10(10), D11(11), D12(12),
								/* face */
    4, D3(5), D6(6), D10(7), D15(8), D21(9), D28(10), D36(11), D45(12)
								/* element */
};
DofType(12, DOF_HB12_, "HB12", HB12_orders, DOF_DG11, FALSE)

static BYTE HB13_orders[] = {
    1,								/* vertex */
    2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13,			/* edge */
    3, D2(4), D3(5), D4(6), D5(7), D8(8), D9(9), D10(10), D11(11), D12(12),
	D13(13),						/* face */
    4, D3(5), D6(6), D10(7), D15(8), D21(9), D28(10), D36(11), D45(12), D55(13)
								/* element */
};
DofType(13, DOF_HB13_, "HB13", HB13_orders, DOF_DG12, FALSE)

static BYTE HB14_orders[] = {
    1,								/* vertex */
    2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14,			/* edge */
    3, D2(4), D3(5), D4(6), D5(7), D8(8), D9(9), D10(10), D11(11), D12(12),
	D13(13), D14(14),					/* face */
    4, D3(5), D6(6), D10(7), D15(8), D21(9), D28(10), D36(11), D45(12),
	D55(13), D66(14)					/* element */
};
DofType(14, DOF_HB14_, "HB14", HB14_orders, DOF_DG13, FALSE)

static BYTE HB15_orders[] = {
    1,								/* vertex */
    2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15,		/* edge */
    3, D2(4), D3(5), D4(6), D5(7), D8(8), D9(9), D10(10), D11(11), D12(12),
	D13(13), D14(14), D15(15),				/* face */
    4, D3(5), D6(6), D10(7), D15(8), D21(9), D28(10), D36(11), D45(12),
	D55(13), D66(14), D78(15)				/* element */
};
DofType(15, DOF_HB15_, "HB15", HB15_orders, DOF_DG14, FALSE)

DOF_TYPE *DOF_HBn[] = {DOF_DG0,
    DOF_HB1,  DOF_HB2,  DOF_HB3,  DOF_HB4,  DOF_HB5,
    DOF_HB6,  DOF_HB7,  DOF_HB8,  DOF_HB9,  DOF_HB10,
    DOF_HB11, DOF_HB12, DOF_HB13, DOF_HB14, DOF_HB15,
    NULL
};

/*------------------------- Hcurl conforming bases -------------------------*/

static const FLOAT *
HC_bas(DOF *dof, SIMPLEX *e, int no0, int no1, const FLOAT *lambda)
{
    static FLOAT *buffers[ORDER_MAX + 1];
    static BOOLEAN initialized = FALSE;

    FLOAT *L;
    const FLOAT (*nabla)[Dim + 1] = (void *)(phgGeomGetJacobian(dof->g, e));

    DOF_TYPE *t = dof->type;
    int order = t->order;
    int nbas = t->nbas;
    FLOAT *bas, *p, *p0, a1, b1, c1, a2, b2, c2, d;
    const FLOAT *phi0, *phi1, *bas_hb;
    DOF hb;
    COORD *coord;
    int i, j, i0, j0, v0, v1, v2, v3, m, n;

    if (!initialized) {
	initialized = TRUE;
	bzero(buffers, sizeof(buffers));
    }

    if (order == 1 && dof->type == DOF_HCn[0])
	order = 0;

    assert(order >= 0 && order <= ORDER_MAX);
    assert(no0 >= t->np_edge * NEdge || no0 % t->np_edge == 0);
    if (no1 <= 0)
	no1 = nbas;
 
    if (no1 <= no0)
	return NULL;

    if ((bas = buffers[order]) == NULL) {
	buffers[order] = phgAlloc(Dim * nbas * sizeof(*bas));
	FreeAtExit(buffers[order]);
	bas = buffers[order];
    }

    p = bas;
    p0 = p + Dim * (no1 - no0);

#if CACHE_LP
    if (L0 == NULL) {
	L0 = phgAlloc(NEdge * sizeof(*L0));
	DL0 = phgAlloc(NEdge * sizeof(*DL0));
	flags = phgAlloc(NEdge * sizeof(*flags));
	FreeAtExit(L0);
	FreeAtExit(DL0);
	FreeAtExit(flags);
    }
    reset_flags = FALSE;
    memset(flags, 0, NEdge * sizeof(*flags));
#else	/* CACHE_LP */
    if (L0 == NULL) {
	L0 = phgAlloc(3 * sizeof(*L0));
	DL0 = phgAlloc(3 * sizeof(*DL0));
	FreeAtExit(L0);
	FreeAtExit(DL0);
    }
#endif	/* CACHE_LP */

    /*------------------------- edge functions --------------------------*/
    n = t->np_edge;
    i0 = no0 / n;
    j0 = no0 % n;
    for (i = i0; i < NEdge && p < p0; i++) {
	GetEdgeVertices(e, i, v0, v1);
	/* first edge basis */
	phi0 = p;
        *(p++) = lambda[v1] * nabla[v0][0] - lambda[v0] * nabla[v1][0];
        *(p++) = lambda[v1] * nabla[v0][1] - lambda[v0] * nabla[v1][1];
        *(p++) = lambda[v1] * nabla[v0][2] - lambda[v0] * nabla[v1][2];
	if (p >= p0)
	    break;
	if (n == 1)
	    continue;
	/* second edge basis */
	phi1 = p;
        *(p++) = lambda[v1] * nabla[v0][0] + lambda[v0] * nabla[v1][0];
        *(p++) = lambda[v1] * nabla[v0][1] + lambda[v0] * nabla[v1][1];
        *(p++) = lambda[v1] * nabla[v0][2] + lambda[v0] * nabla[v1][2];
	if (p >= p0)
	    break;
	if (n == 2)
	    continue;

#if CACHE_LP
	get_L(order - 1, v0, v1, lambda, &L, NULL);
#else	/* CACHE_LP */
	legendre(order - 1, lambda[v1] - lambda[v0], L = L0[0], NULL);
#endif	/* CACHE_LP */
	for (j = (j0 < 2 ? 2 : j0); j < n && p < p0; j++) {
	   /* note: l = j - 1 */
	    b1 = 1. / (FLOAT)j;
	    a1 = (2 * j - 1) * b1;
	    b1 *= j - 1;
	    *(p++) = a1 * L[j - 1] * phi1[0] - b1 * L[j - 2] * phi0[0];
	    *(p++) = a1 * L[j - 1] * phi1[1] - b1 * L[j - 2] * phi0[1];
	    *(p++) = a1 * L[j - 1] * phi1[2] - b1 * L[j - 2] * phi0[2];
	}
	j0 = 0;
    }

    if (p >= p0) {
	reset_flags = TRUE;
	return bas;
    }

    hb.type = DOF_HBn[order];

    /*------------------------- face functions --------------------------*/
    if ((no0 -= NEdge * n) < 0)
	no0 = 0;
    assert(no0 >= t->np_face * NFace || no0 % t->np_face == 0);
    n = t->np_face;
    i0 = no0 / n;
    for (i = i0; i < NFace && p < p0; i++) {
	GetFaceVertices(e, i, v0, v1, v2);
	/* edge-based */
#define EDGE_BASED(v0, v1, v2)						\
	m = NVert + GetEdgeNo(v0, v1) * hb.type->np_edge;		\
	bas_hb = HB_bas(&hb, e, m, m + hb.type->np_edge, lambda);	\
	for (j = 0; j < hb.type->np_edge; j++, bas_hb++) {		\
            *(p++) = *bas_hb * nabla[v2][0];				\
            *(p++) = *bas_hb * nabla[v2][1];				\
            *(p++) = *bas_hb * nabla[v2][2];				\
	}
	EDGE_BASED(v0, v1, v2);
	EDGE_BASED(v0, v2, v1);
	EDGE_BASED(v1, v2, v0);
#undef EDGE_BASED
	if (p >= p0)
	    break;
	if (n == 3 * hb.type->np_edge)
	    continue;
	/* face-based */
	/* get tangent vector of edge (v0, v1) */
	coord = dof->g->verts + e->verts[v1];
	a1 = (*coord)[0];
	b1 = (*coord)[1];
	c1 = (*coord)[2];
	coord = dof->g->verts + e->verts[v2];
	a2 = (*coord)[0];
	b2 = (*coord)[1];
	c2 = (*coord)[2];
	coord = dof->g->verts + e->verts[v0];
	a1 -= (*coord)[0];
	b1 -= (*coord)[1];
	c1 -= (*coord)[2];
	a2 -= (*coord)[0];
	b2 -= (*coord)[1];
	c2 -= (*coord)[2];
	/* normalize tangent vector (not necessary?) */
	d = 1. / Sqrt(a1 * a1 + b1 * b1 + c1 * c1);
	a1 *= d;
	b1 *= d;
	c1 *= d;
	d = 1. / Sqrt(a2 * a2 + b2 * b2 + c2 * c2);
	a2 *= d;
	b2 *= d;
	c2 *= d;
	m = NVert + NEdge * hb.type->np_edge + i * hb.type->np_face;
	bas_hb = HB_bas(&hb, e, m, m + hb.type->np_face, lambda);
	for (j = 0; j < hb.type->np_face; j++, bas_hb++) {
	    *(p++) = *bas_hb * a1;
	    *(p++) = *bas_hb * b1;
	    *(p++) = *bas_hb * c1;

	    *(p++) = *bas_hb * a2;
	    *(p++) = *bas_hb * b2;
	    *(p++) = *bas_hb * c2;
	}
    }

    if (p >= p0) {
	reset_flags = TRUE;
	return bas;
    }

    /*------------------------- element functions --------------------------*/
    if ((no0 -= NFace * n) < 0)
	no0 = 0;
    n = t->np_elem;
    assert(no0 == 0);
    GetElementVertices(e, 0, v0, v1, v2, v3);
    /* face-based */
#define FACE_BASED(v)							\
    m = NVert + NEdge * hb.type->np_edge + v * hb.type->np_face;	\
    bas_hb = HB_bas(&hb, e, m, m + hb.type->np_face, lambda);		\
    for (j = 0; j < hb.type->np_face; j++, bas_hb++) {			\
	*(p++) = *bas_hb * nabla[v][0];					\
	*(p++) = *bas_hb * nabla[v][1];					\
	*(p++) = *bas_hb * nabla[v][2];					\
    }
    FACE_BASED(v0);
    FACE_BASED(v1);
    FACE_BASED(v2);
    FACE_BASED(v3);
#undef FACE_BASED
    if (p >= p0) {
	reset_flags = TRUE;
	return bas;
    }
    /* element-based */
    m = NVert + NEdge * hb.type->np_edge + NFace * hb.type->np_face;
    bas_hb = HB_bas(&hb, e, m, m + hb.type->np_elem, lambda);
    for (j = 0; j < hb.type->np_elem; j++, bas_hb++) {
	*(p++) = *bas_hb;
	*(p++) = 0.;
	*(p++) = 0.;

	*(p++) = 0.;
	*(p++) = *bas_hb;
	*(p++) = 0.;

	*(p++) = 0.;
	*(p++) = 0.;
	*(p++) = *bas_hb;
    }

    assert(p == p0);

    reset_flags = TRUE;
    return bas;
}

static const FLOAT *
HC_grad(DOF *dof, SIMPLEX *e, int no0, int no1, const FLOAT *lambda)
{
    static FLOAT *buffers[ORDER_MAX + 1];
    static BOOLEAN initialized = FALSE;

    FLOAT *L, *DL;
    const FLOAT (*nabla)[Dim + 1] = (void *)(phgGeomGetJacobian(dof->g, e));

    DOF_TYPE *t = dof->type;
    int order = t->order;
    int nbas = t->nbas;
    FLOAT *grad, *p, *p0, a1, b1, c1, a2, b2, c2, d;
    const FLOAT (*dphi0)[Dim + 1], (*dphi1)[Dim + 1], *grad_hb;
    FLOAT phi0[3], phi1[3];
    DOF hb;
    COORD *coord;
    int i, j, i0, j0, v0, v1, v2, v3, m, n;

    if (!initialized) {
	initialized = TRUE;
	bzero(buffers, sizeof(buffers));
    }

    if (order == 1 && dof->type == DOF_HCn[0])
	order = 0;

    assert(order >= 0 && order <= ORDER_MAX);
    assert(no0 >= t->np_edge * NEdge || no0 % t->np_edge == 0);
    if (no1 <= 0)
	no1 = nbas;
 
    if (no1 <= no0)
	return NULL;

    if ((grad = buffers[order]) == NULL) {
	buffers[order] = phgAlloc((Dim + 1) * Dim * nbas * sizeof(*grad));
	FreeAtExit(buffers[order]);
	grad = buffers[order];
    }

    p = grad;
    p0 = p + (Dim + 1) * Dim * (no1 - no0);
    bzero(p, (no1 - no0) * (Dim + 1) * Dim * sizeof(*p));

#if CACHE_LP
    if (L0 == NULL) {
	L0 = phgAlloc(NEdge * sizeof(*L0));
	DL0 = phgAlloc(NEdge * sizeof(*DL0));
	flags = phgAlloc(NEdge * sizeof(*flags));
	FreeAtExit(L0);
	FreeAtExit(DL0);
	FreeAtExit(flags);
    }
    reset_flags = FALSE;
    memset(flags, 0, NEdge * sizeof(*flags));
#else	/* CACHE_LP */
    if (L0 == NULL) {
	L0 = phgAlloc(3 * sizeof(*L0));
	DL0 = phgAlloc(3 * sizeof(*DL0));
	FreeAtExit(L0);
	FreeAtExit(DL0);
    }
#endif	/* CACHE_LP */

    /*------------------------- edge functions --------------------------*/
    n = t->np_edge;
    i0 = no0 / n;
    j0 = no0 % n;
    for (i = i0; i < NEdge && p < p0; i++) {
	GetEdgeVertices(e, i, v0, v1);
	/* first edge basis */
	dphi0 = (void *)p;
        p[v0] = -nabla[v1][0];
        p[v1] =  nabla[v0][0];
	p += Dim + 1;
        p[v0] = -nabla[v1][1];
        p[v1] =  nabla[v0][1];
	p += Dim + 1;
        p[v0] = -nabla[v1][2];
        p[v1] =  nabla[v0][2];
	p += Dim + 1;
	if (p >= p0)
	    break;
	if (n == 1)
	    continue;
	/* second edge basis */
	dphi1 = (void *)p;
        p[v0] = nabla[v1][0];
        p[v1] = nabla[v0][0];
	p += Dim + 1;
        p[v0] = nabla[v1][1];
        p[v1] = nabla[v0][1];
	p += Dim + 1;
        p[v0] = nabla[v1][2];
        p[v1] = nabla[v0][2];
	p += Dim + 1;
	if (p >= p0)
	    break;
	if (n == 2)
	    continue;

	/* need values of the first edge basis functions */
        phi0[0] = lambda[v1] * nabla[v0][0] - lambda[v0] * nabla[v1][0];
        phi0[1] = lambda[v1] * nabla[v0][1] - lambda[v0] * nabla[v1][1];
        phi0[2] = lambda[v1] * nabla[v0][2] - lambda[v0] * nabla[v1][2];

        phi1[0] = lambda[v1] * nabla[v0][0] + lambda[v0] * nabla[v1][0];
        phi1[1] = lambda[v1] * nabla[v0][1] + lambda[v0] * nabla[v1][1];
        phi1[2] = lambda[v1] * nabla[v0][2] + lambda[v0] * nabla[v1][2];

#if CACHE_LP
	get_L(order - 1, v0, v1, lambda, &L, &DL);
#else	/* CACHE_LP */
	legendre(order - 1, lambda[v1] - lambda[v0], L = L0[0], DL = DL0[0]);
#endif	/* CACHE_LP */
	for (j = (j0 < 2 ? 2 : j0) ; j < n && p < p0; j++) {
	   /* note: l = j - 1 */
	    b1 = 1. / (FLOAT)j;
	    a1 = (2 * j - 1) * b1;
	    b1 *= j - 1;
	    p[v0] = a1 * (-DL[j - 1] * phi1[0] + L[j - 1] * dphi1[0][v0]) -
		    b1 * (-DL[j - 2] * phi0[0] + L[j - 2] * dphi0[0][v0]);
	    p[v1] = a1 * ( DL[j - 1] * phi1[0] + L[j - 1] * dphi1[0][v1]) -
		    b1 * ( DL[j - 2] * phi0[0] + L[j - 2] * dphi0[0][v1]);
	    p += Dim + 1;
	    p[v0] = a1 * (-DL[j - 1] * phi1[1] + L[j - 1] * dphi1[1][v0]) -
		    b1 * (-DL[j - 2] * phi0[1] + L[j - 2] * dphi0[1][v0]);
	    p[v1] = a1 * ( DL[j - 1] * phi1[1] + L[j - 1] * dphi1[1][v1]) -
		    b1 * ( DL[j - 2] * phi0[1] + L[j - 2] * dphi0[1][v1]);
	    p += Dim + 1;
	    p[v0] = a1 * (-DL[j - 1] * phi1[2] + L[j - 1] * dphi1[2][v0]) -
		    b1 * (-DL[j - 2] * phi0[2] + L[j - 2] * dphi0[2][v0]);
	    p[v1] = a1 * ( DL[j - 1] * phi1[2] + L[j - 1] * dphi1[2][v1]) -
		    b1 * ( DL[j - 2] * phi0[2] + L[j - 2] * dphi0[2][v1]);
	    p += Dim + 1;
	}
	j0 = 0;
    }

    if (p >= p0) {
	reset_flags = TRUE;
	return grad;
    }

    hb.type = DOF_HBn[order];

    /*------------------------- face functions --------------------------*/
    if ((no0 -= NEdge * n) < 0)
	no0 = 0;
    assert(no0 >= t->np_face * NFace || no0 % t->np_face == 0);
    n = t->np_face;
    i0 = no0 / n;
    for (i = i0; i < NFace && p < p0; i++) {
	GetFaceVertices(e, i, v0, v1, v2);
	/* edge-based */
#define EDGE_BASED(v0, v1, v2)						\
	m = NVert + GetEdgeNo(v0, v1) * hb.type->np_edge;		\
	grad_hb = HB_grad(&hb, e, m, m + hb.type->np_edge, lambda);	\
	for (j = 0; j < hb.type->np_edge; j++, grad_hb += Dim + 1) {	\
            *(p++) = grad_hb[0] * nabla[v2][0];				\
            *(p++) = grad_hb[1] * nabla[v2][0];				\
            *(p++) = grad_hb[2] * nabla[v2][0];				\
            *(p++) = grad_hb[3] * nabla[v2][0];				\
									\
            *(p++) = grad_hb[0] * nabla[v2][1];				\
            *(p++) = grad_hb[1] * nabla[v2][1];				\
            *(p++) = grad_hb[2] * nabla[v2][1];				\
            *(p++) = grad_hb[3] * nabla[v2][1];				\
									\
            *(p++) = grad_hb[0] * nabla[v2][2];				\
            *(p++) = grad_hb[1] * nabla[v2][2];				\
            *(p++) = grad_hb[2] * nabla[v2][2];				\
            *(p++) = grad_hb[3] * nabla[v2][2];				\
	}
	EDGE_BASED(v0, v1, v2);
	EDGE_BASED(v0, v2, v1);
	EDGE_BASED(v1, v2, v0);
#undef EDGE_BASED
	if (p >= p0)
	    break;
	if (n == 3 * hb.type->np_edge)
	    continue;
	/* face-based */
	/* get tangent vector of edge (v0, v1) */
	coord = dof->g->verts + e->verts[v1];
	a1 = (*coord)[0];
	b1 = (*coord)[1];
	c1 = (*coord)[2];
	coord = dof->g->verts + e->verts[v2];
	a2 = (*coord)[0];
	b2 = (*coord)[1];
	c2 = (*coord)[2];
	coord = dof->g->verts + e->verts[v0];
	a1 -= (*coord)[0];
	b1 -= (*coord)[1];
	c1 -= (*coord)[2];
	a2 -= (*coord)[0];
	b2 -= (*coord)[1];
	c2 -= (*coord)[2];
	/* normalize tangent vector (not necessary?) */
	d = 1. / Sqrt(a1 * a1 + b1 * b1 + c1 * c1);
	a1 *= d;
	b1 *= d;
	c1 *= d;
	d = 1. / Sqrt(a2 * a2 + b2 * b2 + c2 * c2);
	a2 *= d;
	b2 *= d;
	c2 *= d;
	m = NVert + NEdge * hb.type->np_edge + i * hb.type->np_face;
	grad_hb = HB_grad(&hb, e, m, m + hb.type->np_face, lambda);
	for (j = 0; j < hb.type->np_face; j++, grad_hb += Dim + 1) {
	    *(p++) = grad_hb[0] * a1;
	    *(p++) = grad_hb[1] * a1;
	    *(p++) = grad_hb[2] * a1;
	    *(p++) = grad_hb[3] * a1;

	    *(p++) = grad_hb[0] * b1;
	    *(p++) = grad_hb[1] * b1;
	    *(p++) = grad_hb[2] * b1;
	    *(p++) = grad_hb[3] * b1;

	    *(p++) = grad_hb[0] * c1;
	    *(p++) = grad_hb[1] * c1;
	    *(p++) = grad_hb[2] * c1;
	    *(p++) = grad_hb[3] * c1;

	    *(p++) = grad_hb[0] * a2;
	    *(p++) = grad_hb[1] * a2;
	    *(p++) = grad_hb[2] * a2;
	    *(p++) = grad_hb[3] * a2;

	    *(p++) = grad_hb[0] * b2;
	    *(p++) = grad_hb[1] * b2;
	    *(p++) = grad_hb[2] * b2;
	    *(p++) = grad_hb[3] * b2;

	    *(p++) = grad_hb[0] * c2;
	    *(p++) = grad_hb[1] * c2;
	    *(p++) = grad_hb[2] * c2;
	    *(p++) = grad_hb[3] * c2;
	}
    }

    if (p >= p0) {
	reset_flags = TRUE;
	return grad;
    }

    /*------------------------- element functions --------------------------*/
    if ((no0 -= NFace * n) < 0)
	no0 = 0;
    n = t->np_elem;
    assert(no0 == 0);
    GetElementVertices(e, 0, v0, v1, v2, v3);
    /* face-based */
#define FACE_BASED(v)							\
    m = NVert + NEdge * hb.type->np_edge + v * hb.type->np_face;	\
    grad_hb = HB_grad(&hb, e, m, m + hb.type->np_face, lambda);		\
    for (j = 0; j < hb.type->np_face; j++, grad_hb += Dim + 1) {	\
	*(p++) = grad_hb[0] * nabla[v][0];				\
	*(p++) = grad_hb[1] * nabla[v][0];				\
	*(p++) = grad_hb[2] * nabla[v][0];				\
	*(p++) = grad_hb[3] * nabla[v][0];				\
									\
	*(p++) = grad_hb[0] * nabla[v][1];				\
	*(p++) = grad_hb[1] * nabla[v][1];				\
	*(p++) = grad_hb[2] * nabla[v][1];				\
	*(p++) = grad_hb[3] * nabla[v][1];				\
									\
	*(p++) = grad_hb[0] * nabla[v][2];				\
	*(p++) = grad_hb[1] * nabla[v][2];				\
	*(p++) = grad_hb[2] * nabla[v][2];				\
	*(p++) = grad_hb[3] * nabla[v][2];				\
    }
    FACE_BASED(v0);
    FACE_BASED(v1);
    FACE_BASED(v2);
    FACE_BASED(v3);
#undef FACE_BASED
    if (p >= p0) {
	reset_flags = TRUE;
	return grad;
    }
    /* element-based */
    m = NVert + NEdge * hb.type->np_edge + NFace * hb.type->np_face;
    grad_hb = HB_grad(&hb, e, m, m + hb.type->np_elem, lambda);
    for (j = 0; j < hb.type->np_elem; j++, grad_hb += Dim + 1) {
	memcpy(p + 0 * (Dim + 1), grad_hb, (Dim + 1) * sizeof(*p));
	p += Dim * (Dim + 1);
	memcpy(p + 1 * (Dim + 1), grad_hb, (Dim + 1) * sizeof(*p));
	p += Dim * (Dim + 1);
	memcpy(p + 2 * (Dim + 1), grad_hb, (Dim + 1) * sizeof(*p));
	p += Dim * (Dim + 1);
    }

    assert(p == p0);

    reset_flags = TRUE;
    return grad;
}

static const FLOAT *
face_normal(GRID *g, SIMPLEX *e, int face)
{
    static FLOAT normal[3 * NFace];
    FLOAT *n = normal + 3 * face;
    FLOAT x0, y0, z0, x1, y1, z1, x2, y2, z2;
    COORD *c;
    int u0, u1, u2;

    if (IsLeaf(e)) {
	memcpy(n, phgGeomGetFaceNormal(g, e, face), 3 * sizeof(*n));
    }
    else {
	GetFaceVertices(e, face, u0, u1, u2);
	x0 = (*(c = g->verts + e->verts[u0]))[0];
	y0 = (*c)[1];
	z0 = (*c)[2];
	x1 = (*(c = g->verts + e->verts[u1]))[0] - x0;
	y1 = (*c)[1] - y0;
	z1 = (*c)[2] - z0;
	x2 = (*(c = g->verts + e->verts[u2]))[0] - x0;
	y2 = (*c)[1] - y0;
	z2 = (*c)[2] - z0;
	n[0] = y1 * z2 - y2 * z1;
	n[1] = z1 * x2 - z2 * x1;
	n[2] = x1 * y2 - x2 * y1;
	/* FIXME: normalize n? */
    }
    /* FIXME: adjust direction of n? */

    return n;
}

static void
HC1_init(DOF *dof, SIMPLEX *e, GTYPE type, int index,
	 DOF_USER_FUNC userfunc, DOF_USER_FUNC_LAMBDA userfunc_lambda,
	 const FLOAT *funcvalues, FLOAT *dofvalues, FLOAT **pdofvalues)
/* computes by L2 projection edge DOF */
{
#if 0	/*********************************************************************/
    GRID *g = dof->g;
    FLOAT lam[] = {0., 0., 0., 0.};
    int i, j, k, n, u0, u1, n0, n1, m = dof->dim;
    FLOAT x0, y0, z0, x1, y1, z1, x2, y2, z2, x, y, z;
    COORD *c;
    QUAD *quad;
    FLOAT *fvals = phgAlloc(Dim * m * sizeof(*fvals));
    FLOAT *temp0, *temp1, *A, *B = dofvalues;
    FLOAT w0, *r;
    const FLOAT *p, *q, *qw, *qp, *b;

    Unused(pdofvalues);

    assert(type == EDGE);
    assert(funcvalues == NULL);
    assert(userfunc_lambda != NULL || userfunc != NULL);

    n = dof->type->np_edge;
    temp0 = phgAlloc(n * sizeof(*temp0));
    temp1 = phgAlloc(m * sizeof(*temp1));
    A = phgCalloc(n * n, sizeof(*A));

    for (i = 0; i < n * m; i++)
	B[i] = 0.0;

    GetEdgeVertices(e, index, u0, u1);
    x0 = (*(c = g->verts + e->verts[u0]))[0];
    y0 = (*c)[1];
    z0 = (*c)[2];
    x1 = (*(c = g->verts + e->verts[u1]))[0];
    y1 = (*c)[1];
    z1 = (*c)[2];
    x2 = x1 - x0;
    y2 = y1 - y0;
    z2 = z1 - z0;

    quad = phgQuadGetQuad1D(2 * dof->type->order);
    qp = quad->points;
    qw = quad->weights;
    /* range of basis functions on the current edge */
    n0 = n * index;
    n1 = n0 + n;
    for (k = 0; k < quad->npoints; k++) {
	lam[u0] = *(qp++);
	lam[u1] = *(qp++);
	if (userfunc_lambda != NULL) {
	    userfunc_lambda(dof, e, -1, lam, fvals);
	}
	else {
	    phgGeomLambda2XYZ(g, e, lam, &x, &y, &z);
	    userfunc(x, y, z, fvals);
	}
	/* inner product of fvals with the tangent vector */
	for (p = fvals, r = temp1, i = 0; i < m; i++, p += 3)
	    *(r++) = x2 * p[0] + y2 * p[1] + z2 * p[2];
	b = dof->type->BasFuncs(dof, e, n0, n1, lam);
	/* inner product of the basis funcs with the tangent vector */
	for (p = b, r = temp0, i = 0; i < n; i++, p += 3)
	    *(r++) = x2 * p[0] + y2 * p[1] + z2 * p[2];
	/* coefficients and RHS */
	w0 = *(qw++);
	for (i = 0, p = temp0; i < n; i++, p++) {
	    for (j = i, q = p; j < n; j++)
		A[i * n + j] += *p * *(q++) * w0;
	    for (j = 0, q = temp1; j < m; j++)
		B[i * m + j] += *p * *(q++) * w0;
	}
    }
    for (i = 1; i < n; i++)
	for (j = 0; j < i; j++)
	    A[i * n + j] = A[j * n + i];
#if 0
printf("---------------------------------------------\n");
for (i = 0; i < n; i++) {
for (j = 0; j < n; j++)
if (Fabs(w0=A[i * n + j])<1e-7) printf("       "); else printf(" %0.2le", (double)w0);
printf(" = %lg\n", (double)B[i]);
}
printf("---------------------------------------------\n");
#endif

    /* solver the linear system of equations */
    if (!phgSolverDenseSolver(n, m, (void *)A, (void *)B))
	phgError(1, "%s:%d, singular system (shouldn't happen).\n",
			__FILE__, __LINE__);
    phgFree(A);
    phgFree(temp0);
    phgFree(temp1);
    phgFree(fvals);
#else	/*********************************************************************/
    /* Note: for edge basis functions, the tangential component are
     * L2 orthogonal on the edge due to the fact that for any j
     *		tau . phi_j = L_j(xi)
     * where tau is the edge vector (v1-v0). The above relation also
     * implies that <tau . phi_j, tau . phi_j>_e is independent of
     * the geometry (e denotes the edge) */
    GRID *g = dof->g;
    FLOAT lam[] = {0., 0., 0., 0.};
    int i, j, k, n, u0, u1, n0, n1, m = dof->dim;
    FLOAT x0, y0, z0, x1, y1, z1, x2, y2, z2, x, y, z;
    COORD *c;
    QUAD *quad;
    FLOAT *fvals = phgAlloc(Dim * m * sizeof(*fvals));
    FLOAT *temp0, *temp1, *B = dofvalues;
    FLOAT w0, *r;
    const FLOAT *p, *q, *qw, *qp, *b;
    static FLOAT *A = NULL;
    static int size_A = 0;

    FreeAtExit(A);
    Unused(pdofvalues);

    assert(type == EDGE);
    assert(funcvalues == NULL);
    assert(userfunc_lambda != NULL || userfunc != NULL);

    n = dof->type->np_edge;
    temp0 = phgAlloc(n * sizeof(*temp0));
    temp1 = phgAlloc(m * sizeof(*temp1));

    if (size_A < n) {
	A = phgRealloc(A, n * sizeof(*A));
	for (i = size_A; i < n; i++)
	    A[i] = 0.;
    }

    for (i = 0; i < n * m; i++)
	B[i] = 0.0;

    GetEdgeVertices(e, index, u0, u1);
    x0 = (*(c = g->verts + e->verts[u0]))[0];
    y0 = (*c)[1];
    z0 = (*c)[2];
    x1 = (*(c = g->verts + e->verts[u1]))[0];
    y1 = (*c)[1];
    z1 = (*c)[2];
    x2 = x1 - x0;
    y2 = y1 - y0;
    z2 = z1 - z0;

    quad = phgQuadGetQuad1D(2 * dof->type->order);
    qp = quad->points;
    qw = quad->weights;
    /* range of basis functions on the current edge */
    n0 = n * index;
    n1 = n0 + n;
    for (k = 0; k < quad->npoints; k++) {
	lam[u0] = *(qp++);
	lam[u1] = *(qp++);
	if (userfunc_lambda != NULL) {
	    userfunc_lambda(dof, e, -1, lam, fvals);
	}
	else {
	    phgGeomLambda2XYZ(g, e, lam, &x, &y, &z);
	    userfunc(x, y, z, fvals);
	}
	/* inner product of fvals with the tangent vector */
	for (p = fvals, r = temp1, i = 0; i < m; i++, p += 3)
	    *(r++) = x2 * p[0] + y2 * p[1] + z2 * p[2];
	b = dof->type->BasFuncs(dof, e, n0, n1, lam);
	/* inner product of the basis funcs with the tangent vector */
	for (p = b, r = temp0, i = 0; i < n; i++, p += 3)
	    *(r++) = x2 * p[0] + y2 * p[1] + z2 * p[2];
	/* coefficients and RHS */
	w0 = *(qw++);
	for (i = 0, p = temp0; i < n; i++, p++) {
	    if (i >= size_A)
		A[i] += *p * *p * w0;
	    for (j = 0, q = temp1; j < m; j++)
		B[i * m + j] += *p * *(q++) * w0;
	}
    }

    size_A = n;
    for (i = 0, p = A; i < n; i++, p++) {
	for (j = 0; j < m; j++)
	    *(B++) /= *p;
    }

    phgFree(temp0);
    phgFree(temp1);
    phgFree(fvals);
#endif	/*********************************************************************/
}

static void
HC2_init(DOF *dof, SIMPLEX *e, GTYPE type, int index,
	 DOF_USER_FUNC userfunc, DOF_USER_FUNC_LAMBDA userfunc_lambda,
	 const FLOAT *funcvalues, FLOAT *dofvalues, FLOAT **pdofvalues)
/* computes by L2 projection face DOF */
{
    GRID *g = dof->g;
    FLOAT lam[] = {0., 0., 0., 0.};
    int i, j, k, n, u0, u1, u2, n0, n1, eno, m = dof->dim;
    FLOAT x, y, z;
    QUAD *quad;
    FLOAT *fvals = phgAlloc(Dim * m * sizeof(*fvals));
    FLOAT *temp, *A, *B = dofvalues;
    FLOAT w0, *r;
    const FLOAT *p, *q, *qw, *qp, *b, *nv;

    assert(type == FACE);
    assert(userfunc_lambda != NULL || userfunc != NULL);
    assert(funcvalues == NULL);

    n = dof->type->np_face;
    temp = phgAlloc(Dim * n * sizeof(*temp));
    A = phgCalloc(n * n, sizeof(*A));

    for (i = 0; i < n * m; i++)
	B[i] = 0.0;

    /* range of basis functions on the current face */
    n0 = dof->type->np_edge * NEdge + n * index;
    n1 = n0 + n;

    GetFaceVertices(e, index, u0, u1, u2);
    nv = face_normal(g, e, index);
    quad = phgQuadGetQuad2D(2 * dof->type->order);
    qp = quad->points;
    qw = quad->weights;
    for (k = 0; k < quad->npoints; k++) {
	lam[u0] = *(qp++);
	lam[u1] = *(qp++);
	lam[u2] = *(qp++);
	if (userfunc_lambda != NULL) {
	    userfunc_lambda(dof, e, -1, lam, fvals);
	}
	else {
	    phgGeomLambda2XYZ(g, e, lam, &x, &y, &z);
	    userfunc(x, y, z, fvals);
	}
	/* evaluate basis functions, since we also need values of the
	 * basis functions on all edges, the range is set to
	 * (0, n1) */
	b = dof->type->BasFuncs(dof, e, 0, n1, lam);
	/* fvals -= contributions from edges. Note that at this stage
	 * dof->data already contains valid data on all edges */
	for (eno = 0, p = b; eno < NEdge; eno++) {
	    /* Note: only need to remove contributions of the
	     * basis functions on the three edges of the face */
	    if (eno != GetEdgeNo(u0,u1) && eno != GetEdgeNo(u0,u2) &&
		eno != GetEdgeNo(u1,u2)) {
		p += dof->type->np_edge * 3;
		continue;
	    }
	    if (pdofvalues == NULL)
		q = DofEdgeData(dof, e->edges[eno]);
	    else
		q = pdofvalues[NVert + eno];
	    for (i = 0; i < dof->type->np_edge; i++, p += 3) {
		for (j = 0, r = fvals; j < m; j++) {
		    w0 = *(q++);
		    *(r++) -= w0 * p[0];
		    *(r++) -= w0 * p[1];
		    *(r++) -= w0 * p[2];
		}
	    }
	}
	/* cross product of fvals with the face normal */
	for (i = 0, r = fvals; i < m; i++) {
	    x = r[1] * nv[2] - r[2] * nv[1];
	    y = r[2] * nv[0] - r[0] * nv[2];
	    z = r[0] * nv[1] - r[1] * nv[0];
	    (*r++) = x;
	    (*r++) = y;
	    (*r++) = z;
	}
	b += 3 * n0;
	/* cross product of basis functions with the face normal */
	for (i = 0, r = temp, p = b; i < n; i++, p += 3) {
	    *(r++) = p[1] * nv[2] - p[2] * nv[1];
	    *(r++) = p[2] * nv[0] - p[0] * nv[2];
	    *(r++) = p[0] * nv[1] - p[1] * nv[0];
	}
	/* coefficients and RHS */
	w0 = *(qw++);
	for (i = 0, p = temp; i < n; i++, p += 3) {
	    for (j = i, q = p; j < n; j++, q += 3)
		A[i * n + j] += (p[0]*q[0] + p[1]*q[1] + p[2]*q[2]) * w0;
	    for (j = 0, q = fvals; j < m; j++, q += 3)
		B[i * m + j] += (p[0]*q[0] + p[1]*q[1] + p[2]*q[2]) * w0;
	}
    }

    for (i = 1; i < n; i++)
	for (j = 0; j < i; j++)
	    A[i * n + j] = A[j * n + i];
#if 0
printf("---------------------------------------------\n");
for (i = 0; i < n; i++) {
for (j = 0; j < n; j++) printf(" %lg",
				(double)(Fabs(A[i * n + j]) < 1e-10 ? 0. : 1.));
printf("\n");
}
printf("---------------------------------------------\n");
printf("\n");
#endif
    /* solver the linear system of equations. */
    if (!phgSolverDenseSolver(n, m, (void *)A, (void *)B))
	phgError(1, "%s:%d, singular system (shouldn't happen).\n",
			__FILE__, __LINE__);

    phgFree(A);
    phgFree(temp);
    phgFree(fvals);
}

static void
HC3_init(DOF *dof, SIMPLEX *e, GTYPE type, int index,
	 DOF_USER_FUNC userfunc, DOF_USER_FUNC_LAMBDA userfunc_lambda,
	 const FLOAT *funcvalues, FLOAT *dofvalues, FLOAT **pdofvalues)
/* computes by L2 projection element DOF */
{
    GRID *g = dof->g;
    int i, j, k, n, u0, u1, u2, u3, n0, n1, no, m = dof->dim;
    QUAD *quad;
    FLOAT *fvals = phgAlloc(Dim * m * sizeof(*fvals));
    FLOAT x, y, z, *A, *B = dofvalues;
    FLOAT w0, *r;
    const FLOAT *p, *q, *qw, *qp, *b;

    assert(type == ELEMENT);
    assert(index == 0);
    assert(userfunc_lambda != NULL || userfunc != NULL);
    assert(funcvalues == NULL);

    n = dof->type->np_elem;
    A = phgCalloc(n * n, sizeof(*A));

    for (i = 0; i < n * m; i++)
	B[i] = 0.0;

    /* range of basis functions in the current element */
    n0 = dof->type->np_edge * NEdge + dof->type->np_face * NFace;
    n1 = n0 + n;

    GetElementVertices(e, index, u0, u1, u2, u3);
    quad = phgQuadGetQuad3D(2 * dof->type->order);
    qp = quad->points;
    qw = quad->weights;
    for (k = 0; k < quad->npoints; k++, qp += 4) {
	if (userfunc_lambda != NULL) {
	    userfunc_lambda(dof, e, -1, qp, fvals);
	}
	else {
	    phgGeomLambda2XYZ(g, e, qp, &x, &y, &z);
	    userfunc(x, y, z, fvals);
	}
	/* evaluate basis functions, since we also need values of the
	 * basis functions on all edges/faces, the range is set to
	 * (0, n1) */
	b = dof->type->BasFuncs(dof, e, 0, n1, qp);
	/* fvals -= contributions from edges. Note that at this stage
	 * dof->data already contains valid data on all edges */
	for (no = 0, p = b; no < NEdge; no++) {
	    if (pdofvalues == NULL)
		q = DofEdgeData(dof, e->edges[no]);
	    else
		q = pdofvalues[NVert + no];
	    for (i = 0; i < dof->type->np_edge; i++, p += 3) {
		for (j = 0, r = fvals; j < m; j++) {
		    w0 = *(q++);
		    *(r++) -= w0 * p[0];
		    *(r++) -= w0 * p[1];
		    *(r++) -= w0 * p[2];
		}
	    }
	}
	/* fvals -= contributions from faces. Note that at this stage
	 * dof->data already contains valid data on all faces */
	for (no = 0; no < NFace; no++) {
	    if (pdofvalues == NULL)
		q = DofFaceData(dof, e->faces[no]);
	    else
		q = pdofvalues[NVert + NEdge + no];
	    for (i = 0; i < dof->type->np_face; i++, p += 3) {
		for (j = 0, r = fvals; j < m; j++) {
		    w0 = *(q++);
		    *(r++) -= w0 * p[0];
		    *(r++) -= w0 * p[1];
		    *(r++) -= w0 * p[2];
		}
	    }
	}
	b += 3 * n0;
	/* coefficients and RHS */
	w0 = *(qw++);
	for (i = 0, p = b; i < n; i++, p += 3) {
	    for (j = i, q = p; j < n; j++, q += 3)
		A[i * n + j] += (p[0] * q[0] + p[1] * q[1] + p[2] * q[2]) * w0;
	    for (j = 0, q = fvals; j < m; j++, q += 3) {
		B[i * m + j] += (p[0] * q[0] + p[1] * q[1] + p[2] * q[2]) * w0;
	    }
	}
    }

    for (i = 1; i < n; i++)
	for (j = 0; j < i; j++)
	    A[i * n + j] = A[j * n + i];
#if 0
printf("---------------------------------------------\n");
for (i = 0; i < n; i++) {
for (j = 0; j < n; j++)
printf(" %012.4le", (double)(Fabs(w0 = A[i * n + j]) < 1e-10 ? 0. : w0));
printf("\n");
}
printf("---------------------------------------------\n");
printf("\n");
#endif
    /* solver the linear system of equations */
    if (!phgSolverDenseSolver(n, m, (void *)A, (void *)B))
	phgError(1, "%s:%d, singular system (shouldn't happen).\n",
			__FILE__, __LINE__);

    phgFree(A);
    phgFree(fvals);
}

static void
HC_init(DOF *dof, SIMPLEX *e, GTYPE type, int index,
	 DOF_USER_FUNC userfunc, DOF_USER_FUNC_LAMBDA userfunc_lambda,
	 const FLOAT *funcvalues, FLOAT *dofvalues, FLOAT **pdofvalues)
{
#if 0 * DEBUG
    static BOOLEAN warned = FALSE;
    if (phgRank == 0 && !warned && dof->type->orders == NULL) {
	warned = TRUE;
	phgWarning("the 'orders' member of \"%s\" is undefined!\n",
			dof->type->name);
    }
#endif

    switch (type) {
	case VERTEX:
	    phgError(1, "%s:%d, unexpected error.\n", __FILE__, __LINE__);
	    break;
	case EDGE:
	    HC1_init(dof, e, type, index, userfunc, userfunc_lambda,
			funcvalues, dofvalues, pdofvalues);
	    break;
	case FACE:
	    HC2_init(dof, e, type, index, userfunc, userfunc_lambda,
			funcvalues, dofvalues, pdofvalues);
	    break;
	case ELEMENT:
	    HC3_init(dof, e, type, index, userfunc, userfunc_lambda,
			funcvalues, dofvalues, pdofvalues);
	    break;
    }
}

/*-------------------------- DOF types HCn --------------------------*/

static HP_INFO hp_info_HC = {DOF_HCn, 0, 15};

#define NBasEdgeHC(p)	(p + 1)
#define NBasFaceHC(p)	(2 * NBasFace(p) + 3 * NBasEdge(p))
#define NBasElemHC(p)	(3 * NBasElem(p) + 4 * NBasFace(p))

#define DofTypeHC(order, type, name, orders, grad)			\
    DOF_TYPE type = {DofCache, name, NULL, orders,			\
			grad, NULL, &hp_info_HC,			\
			phgDofInterpC2FGeneric,				\
			phgDofInterpF2CGeneric,				\
			HC_init, HC_bas, HC_grad,			\
			FALSE, FALSE, -1,				\
			NEdge * NBasEdgeHC(order) +			\
			NFace * NBasFaceHC(order) + NBasElemHC(order),	\
			(order == 0 ? 1 : order), /* order */		\
			-1,		/* continuity */		\
			3,		/* dim */			\
			0, NBasEdgeHC(order),				\
			NBasFaceHC(order), NBasElemHC(order)};

static BYTE HC0_orders[] = {1};
DofTypeHC(0, DOF_HC0_, "HC0", HC0_orders, DOF_DG0)

static BYTE HC1_orders[] = {1, 1};
DofTypeHC(1, DOF_HC1_, "HC1", HC1_orders, DOF_DG0)

static BYTE HC2_orders[] = {1, 1, 2, 2, 2, 2};
DofTypeHC(2, DOF_HC2_, "HC2", HC2_orders, DOF_DG1)

static BYTE HC3_orders[] = {
    1, 1, 2, 3,
    2, 3, 2, 3, 2, 3, 3,
    3, 3, 3, 3};
DofTypeHC(3,  DOF_HC3_, "HC3", HC3_orders, DOF_DG2)

DofTypeHC(4, DOF_HC4_, "HC4", NULL, DOF_DG3)
DofTypeHC(5, DOF_HC5_, "HC5", NULL, DOF_DG4)
DofTypeHC(6, DOF_HC6_, "HC6", NULL, DOF_DG5)
DofTypeHC(7, DOF_HC7_, "HC7", NULL, DOF_DG6)
DofTypeHC(8, DOF_HC8_, "HC8", NULL, DOF_DG7)
DofTypeHC(9, DOF_HC9_, "HC9", NULL, DOF_DG8)
DofTypeHC(10, DOF_HC10_, "HC10", NULL, DOF_DG9)
DofTypeHC(11, DOF_HC11_, "HC11", NULL, DOF_DG10)
DofTypeHC(12, DOF_HC12_, "HC12", NULL, DOF_DG11)
DofTypeHC(13, DOF_HC13_, "HC13", NULL, DOF_DG12)
DofTypeHC(14, DOF_HC14_, "HC14", NULL, DOF_DG13)
DofTypeHC(15, DOF_HC15_, "HC15", NULL, DOF_DG14)

DOF_TYPE *DOF_HCn[] = {
    DOF_HC0, DOF_HC1, DOF_HC2, DOF_HC3, DOF_HC4,
    DOF_HC5, DOF_HC6, DOF_HC7, DOF_HC8, DOF_HC9,
    DOF_HC10, DOF_HC11, DOF_HC12, DOF_HC13, DOF_HC14, DOF_HC15,
    NULL
};

/*------------------------------------------------------------------------*/

#ifdef TEST
static void     
func_u(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
    *value = x * x - y * y + 2. * z * z;
}

int
main(int argc, char *argv[])
{
    GRID *g;
    DOF *u;
    const char *fn = "../test/tetra.dat";

    phgInit(&argc, &argv);
    g = phgNewGrid(-1);
    if (!phgImport(g, fn, FALSE))
	phgError(1, "can't read file \"%s\".\n", fn);
    u = phgDofNew(g, &DOF_HB1_, 1, NULL, func_u);
    phgDofFree(&u);
    phgFreeGrid(&g);
    phgFinalize();
    return 0;
}

#endif	/* TEST */
