/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2008 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: hp.c,v 1.15 2008/03/12 01:59:02 zlb Exp $
 *
 * Support functions for hp-adaptivity */

#include "phg.h"
#include <string.h>

#if 0
# define CMP_ORDER(a,b)	((a) > (b))	/* order = max(orders) */
#else
# define CMP_ORDER(a,b)	((a) < (b))	/* order = min(orders) */
#endif

void
phgHPSetup(DOF *u)
/* Note: order on element e is (temperarily) stored in e->mark,
 * 	 negative order means keep original order */
{
    GRID *g = u->g;
    SIMPLEX *e;
    BYTE min_order, max_order;
    INT i, k, a[3];
    HP_DATA *hp;
    FLOAT *data, *pdata;
    int ii, old_count, new_count;

    hp = phgAlloc(sizeof(*hp));
    if (u->type != NULL) {
	assert(u->hp == NULL);
	if ((hp->info = u->type->hp_info) == NULL)
	    phgError(1, "%s: \"%s\" is not hierarchical, abort.\n",
				__func__, u->type->name);
    }
    else {
	assert(u->hp != NULL);
	hp->info = u->hp->info;
    }

    phgDebug((2, "setting up hierarchical orders.\n"));
    phgDebug((2, "g->nelem = %d.\n", g->nelem));

    min_order = hp->info->min_order;
    max_order = hp->info->max_order;

    hp->edge_order = phgAlloc(g->nedge * sizeof(*hp->edge_order));
    hp->face_order = phgAlloc(g->nface * sizeof(*hp->face_order));
    hp->elem_order = phgAlloc(g->nelem * sizeof(*hp->elem_order));

    hp->edge_index = phgAlloc((g->nedge + 1) * sizeof(*hp->edge_index));
    hp->face_index = phgAlloc((g->nface + 1) * sizeof(*hp->face_index));
    hp->elem_index = phgAlloc((g->nelem + 1) * sizeof(*hp->elem_index));

    /* compute basis orders on edges/faces/elems */

    for (i = 0; i < g->nedge; i++)
	hp->edge_order[i] = (CMP_ORDER(255, 0) ? 0 : 255);
    for (i = 0; i < g->nface; i++)
	hp->face_order[i] = (CMP_ORDER(255, 0) ? 0 : 255);
    for (i = 0; i < g->nelem; i++)
	hp->elem_order[i] = (CMP_ORDER(255, 0) ? 0 : 255);

    ForAllElements(g, e) {
	if (e->mark < 0)
	    e->mark = (u->hp == NULL ? u->type->order :
				       u->hp->elem_order[e->index]);
	assert(e->mark >= min_order && e->mark <= max_order);
	for (ii = 0; ii < NEdge; ii++) {
	    i = e->edges[ii];
	    if (!CMP_ORDER(hp->edge_order[i], e->mark))
		hp->edge_order[i] = e->mark;
	}
	for (ii = 0; ii < NFace; ii++) {
	    i = e->faces[ii];
	    if (!CMP_ORDER(hp->face_order[i], e->mark))
		hp->face_order[i] = e->mark;
	}
	hp->elem_order[e->index] = e->mark;
    }

#if USE_MPI
    if (g->nprocs > 1) {
	/* Note: a dummy MAP is used to build communication pattern */
	static DOF_TYPE type = {DofCache,
	    "dummy", NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	    NULL, NULL, NULL,
	    TRUE, FALSE, -1, NEdge + NFace, 0, -1, -1,
	    0, 1, 1, 0
	};
	DOF *dof = phgDofNew(g, &type, 1, "dummy", DofNoAction);
	MAP *map;
	BYTE *sbuff, *rbuff;
	INT *V2Lmap;

	phgFree(dof->data);	/* don't need DOF data */
	dof->data = NULL;

	map = phgMapCreate(dof, NULL);

	/* build V2Lmap (reverse map of L2Vmap) */
	assert(map->L2Vmap != NULL);
	V2Lmap = phgAlloc(map->localsize * sizeof(*V2Lmap));
	for (i = 0; i < map->localsize; i++)
	    V2Lmap[map->L2Vmap[i]] = i;

	map->cinfo = phgMapCreateCommInfo(
				map->comm,
				map->nprocs,
				map->rank,
				map->nlocal,
				map->localsize,
				map->O2Gmap,
				map->ordering,
				map->partition);
	sbuff = phgAlloc((map->cinfo->ssize + map->cinfo->rsize)
				* sizeof(*sbuff));
	rbuff = sbuff + map->cinfo->ssize;

	/* copy orders on unowned edges/faces to rbuff */
	for (i = 0; i < map->cinfo->rsize; i++) {
	    k = map->nlocal + map->cinfo->rind[i];
	    assert(k < map->localsize);
	    k = V2Lmap[k];
	    if (k < g->nedge) {
		/* edge data */
		rbuff[i] = hp->edge_order[k];
	    }
	    else {
		k -= g->nedge;
		assert(k < g->nface);
		rbuff[i] = hp->face_order[k];
	    }
	}

	/* send orders on unowned edges/faces to their owners */
	phgAlltoallv(rbuff, map->cinfo->rcnts, map->cinfo->rdsps, PHG_MPI_BYTE,
		     sbuff, map->cinfo->scnts, map->cinfo->sdsps, PHG_MPI_BYTE,
		     map->cinfo->comm);

	/* compare with incoming orders */
	for (i = 0; i < map->cinfo->ssize; i++) {
	    k = map->cinfo->sind[i];
	    assert(k < map->nlocal);
	    k = V2Lmap[k];
	    if (k < g->nedge) {
		/* edge data */
		if (!CMP_ORDER(hp->edge_order[k], sbuff[i]))
		    hp->edge_order[k] = sbuff[i];
	    }
	    else {
		k -= g->nedge;
		assert(k < g->nface);
		if (!CMP_ORDER(hp->face_order[k], sbuff[i]))
		    hp->face_order[k] = sbuff[i];
	    }
	}

	/* update incoming orders */
	for (i = 0; i < map->cinfo->ssize; i++) {
	    k = map->cinfo->sind[i];
	    assert(k < map->nlocal);
	    k = V2Lmap[k];
	    if (k < g->nedge) {
		/* edge data */
		sbuff[i] = hp->edge_order[k];
	    }
	    else {
		k -= g->nedge;
		assert(k < g->nface);
		sbuff[i] = hp->face_order[k];
	    }
	}

	/* send back incoming orders */
	phgAlltoallv(sbuff, map->cinfo->scnts, map->cinfo->sdsps, PHG_MPI_BYTE,
		     rbuff, map->cinfo->rcnts, map->cinfo->rdsps, PHG_MPI_BYTE,
		     map->cinfo->comm);

	/* update orders on unowned edges/faces */
	for (i = 0; i < map->cinfo->rsize; i++) {
	    k = map->nlocal + map->cinfo->rind[i];
	    assert(k < map->localsize);
	    k = V2Lmap[k];
	    if (k < g->nedge) {
		/* edge data */
		hp->edge_order[k] = rbuff[i];
	    }
	    else {
		k -= g->nedge;
		assert(k < g->nface);
		hp->face_order[k] = rbuff[i];
	    }
	}

	phgFree(V2Lmap);
	phgFree(sbuff);
	phgMapDestroy(&map);
	phgDofFree(&dof);
    }
#endif	/* USE_MPI */

    /* initialize data pointers */
    hp->edge_index[0] = a[0] = 0;
    for (i = 0; i < g->nedge; i++) {
	if (g->types_edge[i] == UNREFERENCED) {
	    hp->edge_index[i + 1] = hp->edge_index[i];
	    continue;
	}
	k = u->dim * hp->info->types[hp->edge_order[i]]->np_edge;
	hp->edge_index[i + 1] = hp->edge_index[i] + k;
	if (g->types_edge[i] & OWNER)
	    a[0] += k;
    }

    hp->face_index[0] = a[1] = 0;
    for (i = 0; i < g->nface; i++) {
	if (g->types_face[i] == UNREFERENCED) {
	    hp->face_index[i + 1] = hp->face_index[i];
	    continue;
	}
	k = u->dim * hp->info->types[hp->face_order[i]]->np_face;
	hp->face_index[i + 1] = hp->face_index[i] + k;
	if (g->types_face[i] & OWNER)
	    a[1] += k;
    }

    hp->elem_index[0] = a[2] = 0;
    for (i = 0; i < g->nelem; i++) {
	if (g->types_elem[i] == UNREFERENCED) {
	    hp->elem_index[i + 1] = hp->elem_index[i];
	    continue;
	}
	k = u->dim * hp->info->types[hp->elem_order[i]]->np_elem;
	hp->elem_index[i + 1] = hp->elem_index[i] + k;
	assert(g->types_elem[i] & OWNER);
	a[2] += k;
    }

#if USE_MPI
    {
	INT b[3];
	MPI_Allreduce(a, b, 3, PHG_MPI_INT, MPI_SUM, g->comm);
	hp->edge_data_count_global = b[0];
	hp->face_data_count_global = b[1];
	hp->elem_data_count_global = b[2];
    }
#else	/* USE_MPI */
    hp->edge_data_count_global = a[0];
    hp->face_data_count_global = a[1];
    hp->elem_data_count_global = a[2];
#endif	/* USE_MPI */

    /* set up data */
    pdata = data = phgCalloc(g->nvert * u->count_vert +
			     hp->edge_index[g->nedge] +
			     hp->face_index[g->nface] +
			     hp->elem_index[g->nelem], sizeof(*data));

    /*
     * FIXME: the following relies on orthogonality of bases which may not hold
     */

    /* copy vertex data */
    if ((k = DofGetVertexDataCount(u)) > 0) {
	memcpy(pdata, u->data, k * sizeof(*data));
	pdata += k;
    }

    /* set up edge data */
    for (i = 0; i < g->nedge; i++) {
	old_count = DofGetDataCountOnEdge(u, i);
	new_count = hp->edge_index[i + 1] - hp->edge_index[i];
	ii = (old_count < new_count ? old_count : new_count);
	if (ii > 0)
	    memcpy(pdata, DofEdgeData(u, i), ii * sizeof(*pdata));
	pdata += new_count;
    }

    /* set up face data */
    for (i = 0; i < g->nface; i++) {
	old_count = DofGetDataCountOnFace(u, i);
	new_count = hp->face_index[i + 1] - hp->face_index[i];
	ii = (old_count < new_count ? old_count : new_count);
	if (ii > 0)
	    memcpy(pdata, DofFaceData(u, i), ii * sizeof(*pdata));
	pdata += new_count;
    }

    /* set up element data */
    for (i = 0; i < g->nelem; i++) {
	old_count = DofGetDataCountOnElement(u, i);
	new_count = hp->elem_index[i + 1] - hp->elem_index[i];
	ii = (old_count < new_count ? old_count : new_count);
	if (ii > 0)
	    memcpy(pdata, DofElementData(u, i), ii * sizeof(*pdata));
	pdata += new_count;
    }

    /* update members of u */
    phgFree(u->data);
    u->data = data;
    u->data_edge = u->data_vert + g->nvert * u->count_vert;
    u->data_face = u->data_edge + hp->edge_index[g->nedge];
    u->data_elem = u->data_face + hp->face_index[g->nface];
    if (u->hp != NULL) {
	phgFree(u->hp->edge_order);
	phgFree(u->hp->face_order);
	phgFree(u->hp->elem_order);
	phgFree(u->hp->edge_index);
	phgFree(u->hp->face_index);
	phgFree(u->hp->elem_index);
	phgFree(u->hp);
    }
    u->hp = hp;
    u->type = NULL;	/* accessing u->type will produce segfault */
}
