/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2008 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: solver-petsc.c,v 1.51 2008/01/14 03:22:23 zlb Exp $ */

/* The PETSc interface for PHG */

#include "phg.h"

#if USE_PETSC

#include <petscksp.h>
#include <stdlib.h>

#define Init	NULL

/* for compatibility with PETSc <= 2.2.0 */
#if PETSC_VERSION_MAJOR < 2 || \
    (PETSC_VERSION_MAJOR == 2 && (PETSC_VERSION_MINOR < 2 || \
        (PETSC_VERSION_MINOR == 2 && PETSC_VERSION_SUBMINOR == 0)))
typedef int PetscInt;
typedef int PetscMPIInt;
typedef int PetscErrorCode;
#endif

typedef struct {
    Vec B;	/* B=RHS */
    Vec U;	/* U=solution */
    Mat A;	/* A=matrix */
    KSP ksp;
    BOOLEAN assembly_rhs;
    BOOLEAN assembly_mat;
} OEM_DATA;

/* convenience macro */
#define _t	((OEM_DATA *)solver->oem_data)

static PetscErrorCode code;

static int
RegisterOptions(void)
{
    phgOptionsRegisterTitle("\nPETSc options:",
				"Use '-oem_options \"PETSc options\"' to pass "
				"options to PETSc. "
				"Please check PETSc's manual for "
				"options which can be used with PETSc, "
				"or use '-oem_options \"-help\"' to show "
				"available PETSc options. For example: "
				"'-oem_options \"-ksp_type cg -pc_type none\"'",
				"petsc");

    return 0;
}

static int
Initialize(int *argc, char ***argv)
{
    static char help[] = "\nThe PETSc interface for PHG.\n\n";

    /* Hacky: also initialize SLEPc here */
#if USE_SLEPC
    EXTERN PetscErrorCode SlepcInitialize(int*, char***, char[], const char[]);
    code = SlepcInitialize(argc, argv, NULL, help); CHKERRQ(code);
#else
    code = PetscInitialize(argc, argv, NULL, help); CHKERRQ(code);
#endif

    /* print information about the solver */
    if (phgRank == 0) {
	PetscTruth flg;

#if 0
	PetscTruth flg1;
	char name[32], name1[32];
	PetscOptionsGetString(NULL, "-ksp_type", name, sizeof(name), &flg);
	PetscOptionsGetString(NULL, "-pc_type", name1, sizeof(name1), &flg1);
	phgPrintf("KSP method = %s, precoditioner = %s\n",
		flg ? name : "default", flg1 ? name1 : "default");
#endif

	/* check for obsolete PETSc options */
	PetscOptionsHasName(PETSC_NULL,"-ksp_max_it", &flg);
	if (flg)
	    phgPrintf("WARNING: PETSc option \"-ksp_max_it\" is obsolete, "
		      "use PHG option \"-solver_max_it\" instead.\n");

	PetscOptionsHasName(PETSC_NULL,"-ksp_rtol", &flg);
	if (flg)
	    phgPrintf("WARNING: PETSc option \"-ksp_rtol\" is obsolete, "
		      "use PHG option \"-solver_rtol\" instead.\n");
    }

    return 0;
}

static int
Finalize(void)
{
#if USE_SLEPC
    EXTERN PetscErrorCode SlepcFinalize(void);
    code = SlepcFinalize(); CHKERRQ(code);
#else
    code = PetscFinalize(); CHKERRQ(code);
#endif

    return 0;
}

static int
Create(SOLVER *solver)
{
    MAP *map = solver->mat->rmap;
    PetscInt N = map->nglobal, Nlocal = map->nlocal;
    PetscInt prealloc, *d_nnz, *o_nnz;

    assert(solver->mat->assembled);
    assert(solver->mat->type == PHG_UNPACKED);
    assert(solver->mat->type != PHG_MATRIX_FREE);

    solver->oem_data = phgAlloc(sizeof(OEM_DATA));
    _t->B = PETSC_NULL;
    _t->U = PETSC_NULL;
    _t->A = PETSC_NULL;
    _t->ksp = PETSC_NULL;
    _t->assembly_rhs = FALSE;
    _t->assembly_mat = FALSE;

    {
	PetscInt i, j, d, o;
	MAT_ROW *row;

	prealloc = PETSC_DECIDE;
	d_nnz = phgAlloc(Nlocal * sizeof(*d_nnz));
	o_nnz = phgAlloc(Nlocal * sizeof(*o_nnz));
	row = solver->mat->rows;
	for (i = 0; i < Nlocal; i++, row++) {
	    d = o = 0;
	    for (j = 0; j < row->ncols; j++)
		(row->cols[j] < Nlocal) ? d++ : o++;
	    d_nnz[i] = d;
	    o_nnz[i] = o;
	}
    }

    code = MatCreateMPIAIJ(map->comm, Nlocal, Nlocal, N, N,
				prealloc, d_nnz, prealloc, o_nnz, &_t->A);
    CHKERRQ(code);
    code = MatSetFromOptions(_t->A); CHKERRQ(code);
    code = VecCreate(map->comm, &_t->U); CHKERRQ(code);
    code = VecSetSizes(_t->U, Nlocal, N); CHKERRQ(code);
    code = VecSetFromOptions(_t->U); CHKERRQ(code);
    code = VecDuplicate(_t->U, &_t->B); CHKERRQ(code);

    phgFree(d_nnz);
    phgFree(o_nnz);

    return 0;
}

static int
Destroy(SOLVER *solver)
{
    if (solver->oem_data == NULL)
	return 0;

    if (_t->ksp != PETSC_NULL)
	KSPDestroy(_t->ksp);
    if (_t->A != PETSC_NULL)
	MatDestroy(_t->A);
    if (_t->B != PETSC_NULL)
	VecDestroy(_t->B);
    if (_t->U != PETSC_NULL)
	VecDestroy(_t->U);

    phgFree(solver->oem_data);
    solver->oem_data = NULL;

    return 0;
}

static int
AddMatrixEntries(SOLVER *solver, INT nrows, INT *rows, INT ncols, INT *cols,
			FLOAT *values)
{
    if (nrows == 0 || ncols == 0)
	return 0;

    if (_t->A == PETSC_NULL)
	phgError(1, "PETSc AddMatrixEntries: invalid matrix!\n");

    if (sizeof(INT) != sizeof(PetscInt))
	phgError(1, "%s:%d: unimplemented.\n", __FILE__, __LINE__);

    if (sizeof(FLOAT) == sizeof(PetscScalar)) {
	code = MatSetValues(_t->A, nrows, rows, ncols, cols,
				(PetscScalar *)values, INSERT_VALUES);
    }
    else {
	PetscInt i, n = nrows * ncols;
	PetscScalar *buffer = phgAlloc(n * sizeof(*buffer));
	for (i = 0; i < n; i++)
	    buffer[i] = (PetscScalar)values[i];
	code = MatSetValues(_t->A, nrows, rows, ncols, cols,
				buffer, INSERT_VALUES);
	phgFree(buffer);
    }
    CHKERRQ(code);
    _t->assembly_mat = TRUE;

    return 0;
}

static int
AddRHSEntries(SOLVER *solver, INT n, INT *ni, FLOAT *values)
{
    if (n == 0)
	return 0;

    if (_t->B == PETSC_NULL)
	phgError(1, "PETSc AddRHSEntries: invalid vector!\n");

    if (sizeof(INT) != sizeof(PetscInt))
	phgError(1, "%s:%d: unimplemented.\n", __FILE__, __LINE__);

    if (sizeof(FLOAT) == sizeof(PetscScalar)) {
	code = VecSetValues(_t->B, n, ni, (PetscScalar *)values, INSERT_VALUES);
    }
    else {
	PetscInt i;
	PetscScalar *buffer = phgAlloc(n * sizeof(*buffer));
	for (i = 0; i < n; i++)
	    buffer[i] = (PetscScalar)values[i];
	code = VecSetValues(_t->B, n, ni, buffer, INSERT_VALUES);
	phgFree(buffer);
    }
    CHKERRQ(code);
    _t->assembly_rhs = TRUE;

    return 0;
}
 
static int
Assemble(SOLVER *solver)
{
    if (_t->assembly_rhs) {
	code = VecAssemblyBegin(_t->B); CHKERRQ(code);
	code = VecAssemblyEnd(_t->B); CHKERRQ(code);
	_t->assembly_rhs = FALSE;
    }

    if (_t->assembly_mat) {
	code = MatAssemblyBegin(_t->A, MAT_FINAL_ASSEMBLY); CHKERRQ(code);
	code = MatAssemblyEnd(_t->A, MAT_FINAL_ASSEMBLY); CHKERRQ(code);

	code = KSPCreate(solver->mat->rmap->comm, &_t->ksp); CHKERRQ(code);
	code = KSPSetOperators(_t->ksp, _t->A, _t->A,
				DIFFERENT_NONZERO_PATTERN); CHKERRQ(code);
	code = KSPSetFromOptions(_t->ksp); CHKERRQ(code);
	code = KSPSetInitialGuessNonzero(_t->ksp, PETSC_TRUE); CHKERRQ(code);

	_t->assembly_mat = FALSE;
    }
    
#if 0
    phgPrintf("The RHS:\n");
    PetscViewerSetFormat(PETSC_VIEWER_STDOUT_SELF, PETSC_VIEWER_ASCII_MATLAB);
    VecView(_t->B, PETSC_VIEWER_STDOUT_SELF);
    phgPrintf("The matrix:\n");
    PetscViewerSetFormat(PETSC_VIEWER_STDOUT_SELF, PETSC_VIEWER_ASCII_MATLAB);
    MatView(_t->A, PETSC_VIEWER_STDOUT_SELF);
#endif

    return 0;
}

static int
set_initial_solution(SOLVER *solver, VEC *x)
{
    PetscScalar *vec;
    PetscInt size;
    INT i;

    if (_t->U == PETSC_NULL)
	phgError(1, "PETSc CopySolution: invalid PETSc vector.\n");

    code = VecGetLocalSize(_t->U, &size); CHKERRQ(code);
    code = VecGetArray(_t->U, &vec); CHKERRQ(code);
    for (i = 0; i < size; i++)
	vec[i] = x->data[i];
    code = VecRestoreArray(_t->U, &vec); CHKERRQ(code);

    return 0;
}

static int
copy_solution(SOLVER *solver, VEC *x)
/* copy solution from PETSc Vec 'U' to local DOFs */
{
    PetscScalar *vec;
    PetscInt size;
    INT i;

    if (_t->U == PETSC_NULL)
	phgError(1, "PETSc CopySolution: invalid PETSc vector.\n");

    code = VecGetLocalSize(_t->U, &size); CHKERRQ(code);
    code = VecGetArray(_t->U, &vec); CHKERRQ(code);
    for (i = 0; i < size; i++)
	x->data[i] = vec[i];
    code = VecRestoreArray(_t->U, &vec); CHKERRQ(code);

    return 0;
}

PetscErrorCode ShellPreconditioner(void *ctx,Vec x, Vec y)
{
    SOLVER_PC *pc=((SOLVER*)ctx)->pc;
    PetscScalar *vec_x,*vec_y;
    VEC *x_phg,*y_phg;
    INT i, verb = phgVerbosity;

    /* copy petsc-vec to phg-vec */
    code = VecGetArray(x, &vec_x); CHKERRQ(code);
    code = VecGetArray(y, &vec_y); CHKERRQ(code);
    if (sizeof(PetscScalar) == sizeof(FLOAT)) {
	x_phg=phgMapCreateVec(((SOLVER*)ctx)->mat->rmap, 0);
	y_phg=phgMapCreateVec(((SOLVER*)ctx)->mat->rmap, 0);
	x_phg->data = (FLOAT *)vec_x;
	y_phg->data = (FLOAT *)vec_y;
	x_phg->nvec=1;
	y_phg->nvec=1;
    }
    else {
	x_phg=phgMapCreateVec(((SOLVER*)ctx)->mat->rmap, 1);
	y_phg=phgMapCreateVec(((SOLVER*)ctx)->mat->rmap, 1);
	for (i = 0; i < x_phg->map->nlocal; i++) {
	    x_phg->data[i] = (FLOAT)vec_x[i];
	    y_phg->data[i] = (FLOAT)vec_y[i];
	}
    }

    /*  Precond Proc*/
    if (phgVerbosity > 0)
	phgVerbosity--;
    pc->pc_proc(pc->pc_solver, x_phg, &y_phg);
    phgVerbosity = verb;

    /* copy phg-cvec to petsc-vec */
    if (sizeof(PetscScalar) == sizeof(FLOAT)) {
	x_phg->data = NULL;
	y_phg->data = NULL;
    }
    else {
	for (i = 0; i < x_phg->map->nlocal; i++) {
	    vec_x[i] = (PetscScalar)x_phg->data[i];
	    vec_y[i] = (PetscScalar)y_phg->data[i];
	}
    }
    code = VecRestoreArray(x, &vec_x); CHKERRQ(code);
    code = VecRestoreArray(y, &vec_y); CHKERRQ(code);
    phgVecDestroy(&x_phg);
    phgVecDestroy(&y_phg);
    return code;
}

static int
Solve(SOLVER *solver, VEC *x, BOOLEAN destroy)
{
    PetscInt nits;
    PetscReal residual;
    PC  pc;

    if (_t->A == PETSC_NULL ||
	_t->B == PETSC_NULL ||
	_t->U == PETSC_NULL)
	phgError(1, "PETSc Solve: invalid linear system.\n");

    code = Assemble(solver); CHKERRQ(code);

    set_initial_solution(solver, x);

#if 0
    {
	PetscReal norma, normb, normu;
	MatNorm(_t->A, NORM_FROBENIUS, &norma);
	VecNorm(_t->B, NORM_2, &normb);
	VecNorm(_t->U, NORM_2, &normu);
	phgPrintf("\nNorm(A) = %lg, Norm(b) = %lg, Norm(x0) = %lg\n",
			(double)norma, (double)normb, (double)normu);
    }
#endif

    code = KSPSetTolerances(_t->ksp, solver->rtol, solver->atol,
			    PETSC_DEFAULT, solver->maxit);
    CHKERRQ(code);
   /* user-defined prcond */ 
    if(solver->pc!=NULL){
      KSPGetPC(_t->ksp,&pc);
      PCSetType(pc,PCSHELL);
#if PETSC_VERSION_MAJOR < 2 || \
    (PETSC_VERSION_MAJOR == 2 && PETSC_VERSION_MINOR < 2)
      PCShellSetApply(pc,ShellPreconditioner,(void*)solver);
#else
      PCShellSetContext(pc,(void*)solver);
      PCShellSetApply(pc,ShellPreconditioner);
#endif
    }

#if PETSC_VERSION_MAJOR < 2 || \
    (PETSC_VERSION_MAJOR == 2 && (PETSC_VERSION_MINOR < 2 || \
     (PETSC_VERSION_MINOR == 2 && PETSC_VERSION_SUBMINOR < 1)))
    code = KSPSetRhs(_t->ksp, _t->B); CHKERRQ(code);
    code = KSPSetSolution(_t->ksp, _t->U); CHKERRQ(code);
    code = KSPSolve(_t->ksp); CHKERRQ(code);
#else
    code = KSPSolve(_t->ksp, _t->B, _t->U); CHKERRQ(code);
#endif
    code = KSPGetIterationNumber(_t->ksp, &nits); CHKERRQ(code);
    code = KSPGetResidualNorm(_t->ksp, &residual); CHKERRQ(code);

    if (destroy) {
	KSPDestroy(_t->ksp);
	_t->ksp = PETSC_NULL;
	MatDestroy(_t->A);
	_t->A = PETSC_NULL;
	VecDestroy(_t->B);
	_t->B = PETSC_NULL;
    }

    copy_solution(solver, x);

    if (destroy) {
	VecDestroy(_t->U);
	_t->U = PETSC_NULL;
    }

    solver->residual = residual;
    solver->nits = nits;

    return nits;
}

/*--------------------------------------------------------------------*/

OEM_SOLVER phgSolverPetsc_ = {
    "PETSc", RegisterOptions,
    Initialize, Finalize, Init, Create, Destroy, AddMatrixEntries,
    AddRHSEntries, Assemble, Solve
};

#endif		/* USE_PETSC */
