/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2008 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: solver-phg.c,v 1.38 2008/03/04 02:17:38 zlb Exp $ */

/* The pcg solver.  */

#include "phg.h"
#include "solver-phg.h"
#include <string.h>
#include <strings.h>
#include <math.h>

#define TRUNK_SIZE 8

#define Initialize		NULL
#define Finalize		NULL

static void void_pc(SOLVER *pc_solver, VEC *b, VEC **x);
static void jacobi_pc(SOLVER *pc_solver, VEC *b, VEC **x);

static const char *pc_names[] = {
    "none", "jacobi", NULL
};

static const PC_PROC pc_list[] = {
    void_pc, jacobi_pc, NULL
};

static int pcg_pc_id = 1;	/* default to Jacobi PC */
static int gmres_pc_id = 1;	/* default to Jacobi PC */
static int gmres_restart = 10;

typedef struct {
    SOLVER_PC	*pc;
    BOOLEAN	assembled;
    BOOLEAN	warned;		/* used to warn non SPD matrix only once */
    int		pc_id;
    int		gmres_restart;
} OEM_DATA;

#define _t      ((OEM_DATA *)solver->oem_data)

void 
phgSolverGMRESSetRestart(SOLVER *solver, int restart)
{
    _t->gmres_restart = restart;
}

int
phgSolverJacobi(MAT *A, VEC *b, VEC **px, int max_its, FLOAT rtol,
		FLOAT *res)
{
    VEC *x, *r;
    FLOAT residual = 1.0, rb_norm;
    INT i, nits;

    if (*px == NULL || b == NULL)
	phgError(1, "%s:%d invalid vector handler\n", __FILE__, __LINE__);

    if (A->rows == NULL)
	return 0;

    assert(b->nvec == 1);

    x = *px;
    r = phgMapCreateVec(b->map, 1);

    if (A->diag == NULL)
	phgMatSetupDiagonal(A);

    nits = 0;
    if ((rb_norm = phgVecNorm2(b, 0, NULL)) == 0.0)
	rb_norm = 1.0;
    else
	rb_norm = 1.0 / rb_norm;
    while (residual >= rtol && nits < max_its) {
	phgVecCopy(b, &r);
	phgMatVec(MAT_OP_N, -1.0, A, x, 1.0, &r);
	for (i = 0; i < A->cmap->nlocal; i++)
	    r->data[i] /= A->diag[i];
	nits++;
	residual = phgVecNorm2(r, 0, NULL) * rb_norm;
	phgVecAXPBY(1.0, r, 1.0, &x);
    }			/* end of while-loop */
    phgVecDestroy(&r);
    if (res != NULL)
	*res = residual;

    return nits;
}

static void
default_pc_proc(SOLVER *pc_solver, VEC *b, VEC **x)
{
    FLOAT *old_rhs;
    
    phgSolverAssemble(pc_solver);      /* assemble matrix and RHS */
	
    if (pc_solver->mat->rmap->nlocal != b->map->nlocal) {
	phgPrintf("Invalid PC!\n");
	return;
    }

    /* updating RHS */  
    old_rhs = pc_solver->rhs->data;
    pc_solver->rhs->data = b->data;
    pc_solver->rhs->assembled=TRUE;
    pc_solver->rhs_updated=TRUE;
#if 0
    phgVecCopy(pc_solver->rhs, x);
#else
    bzero((*x)->data, (*x)->map->nlocal * sizeof(*(*x)->data));
#endif
    /* solving M x = b */
    phgSolverVecSolve(pc_solver, FALSE, *x);
    pc_solver->rhs->data = old_rhs;

    return;  
}

static void
void_pc(SOLVER *pc_solver, VEC *b, VEC **x)
{
    phgVecCopy(b, x);
    pc_solver->nits = 0;

    return;
}

static void
jacobi_pc(SOLVER *pc_solver, VEC *b, VEC **x)
{
    MAT *A;
    VEC *v;
    FLOAT *rhs, *p;
    FLOAT value;
    INT i;

    phgSolverAssemble(pc_solver);

#if 0
    /* test the Jacobi solver */
    bzero((*u)->data, B->map->nlocal * sizeof(*(*u)->data));
    /*phgVecCopy(b, x);*/
    phgSolverJacobi(pc_solver->mat, b, x, 1, 0.0, NULL);
    return;
#endif

    A = pc_solver->mat;
    if (A->diag == NULL)
	phgMatSetupDiagonal(A);

    v = *x;
    p = v->data;
    rhs = b->data;
    for (i = 0; i < A->rmap->nlocal; i++, rhs++, p++) {
	if ((value = A->diag[i]) == 0.0)
	    phgError(1, "%s:%d, zero diagonal entry.\n", __FILE__, __LINE__);
	*p = *rhs / value;
    }

    pc_solver->nits = 1;

    return;
}

/*---------------------------------------------------------------------------*/

void
phgSolverSetPC(SOLVER *solver, SOLVER *pc_solver, PC_PROC pc_proc)
{
   SOLVER_PC *pc;

   assert(solver != NULL);

   if (solver->pc == NULL)
	solver->pc = phgAlloc(sizeof(SOLVER_PC));
   pc=((SOLVER_PC *)(solver->pc));
   pc->pc_solver = pc_solver;

    if (pc_proc == NULL) {
        if (solver == pc->pc_solver) {
          pc->pc_proc = pc_list[_t->pc_id];
          phgWarning("Use the %s-pc as PC-Solver's preconditioner \n",
			pc_names[_t->pc_id]);
        } else {
           pc->pc_proc = default_pc_proc;
        }
    } else {
   	pc->pc_proc = pc_proc;
    }

    return;
}
/* -------------------------------------------------------------*/

static int
PCGRegisterOptions(void)
{
    phgOptionsRegisterTitle("\nBuilt-in PCG solver options:", NULL, "pcg");
    phgOptionsRegisterKeyword("pcg_pc_type", "Preconditioner for built-in PCG",
				pc_names, &pcg_pc_id);
    return 0;
}

static int
GMRESRegisterOptions(void)
{
    phgOptionsRegisterTitle("\nBuilt-in GMRES solver options:", NULL, "gmres");
    phgOptionsRegisterKeyword("gmres_pc_type",
			      "Preconditioner for built-in GMRES",
			      pc_names, &gmres_pc_id);
    phgOptionsRegisterInt("gmres_restart",
			  "Restart steps for built-in GMRES",
			  &gmres_restart);
    return 0;
}

int
phgSolverPCG__(MAT *A, VEC *b, VEC **x_ptr, SOLVER_PC *pc, int max_its,
	     FLOAT rtol, FLOAT atol, FLOAT *res, BOOLEAN *warned)
{
    int nits;
    FLOAT residual = 0., rb_norm, tol;
    FLOAT rz_1 = 0.0, rz_2 = 0.0, alpha;
    VEC *r = NULL, *r_1 = NULL, *z = NULL, *Ap = NULL;
    VEC *p = NULL, *p_1 = NULL, *x;

    assert(b->nvec == 1);
    x = *x_ptr;
    nits = 0;
    if (IsZero(rb_norm = phgVecNorm2(b, 0, NULL))) {
	bzero(x->data, sizeof(*x->data) * x->map->nlocal);
	return 0;
    }

    r = phgMapCreateVec(x->map, 1);
    r_1 = phgMapCreateVec(x->map, 1);
    z = phgMapCreateVec(x->map, 1);
    Ap = phgMapCreateVec(x->map, 1);
    p = phgMapCreateVec(x->map, 1);
    p_1 = phgMapCreateVec(x->map, 1);

    /* r = b - A * _t_U  */
    if (!IsZero(phgVecNorm1(x, 0, NULL) / A->cmap->nglobal)) {
	phgMatVec(MAT_OP_N, 1.0, A, x, 0.0, &r_1);
	phgVecCopy(r_1, &r);
	phgVecAXPBY(1.0, b, -1.0, &r);
    }
    else {
	phgVecCopy(b, &r);
    }

    residual = phgVecNorm2(r, 0, NULL);
    tol=rtol*rb_norm;
    if(rtol*rb_norm < atol)
       tol=atol;
    while (residual >= tol && nits < max_its) {
	INT verb = phgVerbosity;
	if (phgVerbosity > 0)
	    phgVerbosity--;
	if (pc != NULL)
	    pc->pc_proc(pc->pc_solver, r, &z);
	else
	    void_pc(NULL, r, &z);
	phgVerbosity = verb;
	nits++;
	if (nits == 1) {
	    phgVecCopy(z, &p);
	    rz_1 = phgVecDot(r, 0, z, 0, NULL);
	}
	else {
	    rz_2 = rz_1;
	    rz_1 = phgVecDot(r, 0, z, 0, NULL);
	    phgVecCopy(p, &p_1);
	    if (rz_2 == 0.0) {
		/*phgError(1, "%s:%d, the matrix may not be SPD.\n",
						__FILE__, __LINE__);*/
		if (warned == NULL || !*warned) {
		    phgPrintf("%s:%d, Warning! the matrix may not be SPD.\n",
				__FILE__, __LINE__);
		    if (warned != NULL)
			*warned = TRUE;
		}
		break;
	    }
	    phgVecAXPBY(1.0, z, rz_1 / rz_2, &p);
	}
	phgMatVec(MAT_OP_N, 1.0, A, p, 0.0, &Ap);
	if ((alpha = phgVecDot(p, 0, Ap, 0, NULL)) == 0.0) {
	    /*phgError(1, "%s:%d, the matrix may not be SPD.\n",
						__FILE__, __LINE__);*/
	    if (warned == NULL || !*warned) {
		phgPrintf("%s:%d, Warning! the matrix may not be SPD.\n",
			  __FILE__, __LINE__);
		if (warned != NULL)
		    *warned = TRUE;
	    }
	    break;
	}
	alpha = rz_1 / alpha;
	phgVecAXPBY(alpha, p, 1.0, &x);
	phgVecCopy(r, &r_1);
	phgVecAXPBY(-1.0 * alpha, Ap, 1.0, &r);
	residual = phgVecNorm2(r, 0, NULL);
	if (phgVerbosity > 0) {
	    phgPrintf("  PCG iteration %d, residual = %le",
				nits, (double)residual);
	    if (pc != NULL /*&& pc->pc_solver != solver*/
		&& pc->pc_solver->nits > 0)
		phgPrintf(" (%d iteration%s in PC)\n", pc->pc_solver->nits,
				pc->pc_solver->nits > 1 ? "s" : "");
	    else
		phgPrintf("\n");
	}
    }

    phgVecDestroy(&r);
    phgVecDestroy(&r_1);
    phgVecDestroy(&z);
    phgVecDestroy(&Ap);
    phgVecDestroy(&p_1);
    phgVecDestroy(&p);

    if (res != NULL)
	*res = residual;

    return nits;
}

int
phgSolverGMRES__(MAT *A, VEC *b, VEC **x_ptr, SOLVER_PC *pc, int restart,
		 int max_its, FLOAT rtol, FLOAT atol, FLOAT *res)
    /*
     *  for details to the algorithm GMRES(m)-Aya see
     *
     *  R. Barrett, M. Berry, T. F. Chan, J. Demmel, J. Donato, J. Dongarra,
     *  V. Eijkhout, R. Pozo, Ch. Romine, H. van der Vorst:
     *  Templates for the Solution of Linear Systems: Building Blocks
     *  for Iterative Solvers;
     *  SIAM, Philadelphia, 1994
     *
     */
{
    int nits,i,j,k;
    BOOLEAN break_flag=FALSE,false_conv=FALSE;
    INT nlocal, nglobal,verb;
    FLOAT residual = 0.,r_norm,b_norm,tol, v_norm;
    FLOAT tmp,h1,h2;
    FLOAT **H=NULL,*h_data=NULL,**HH=NULL,*hh_data=NULL;
    FLOAT *y=NULL,*s=NULL, *c1=NULL,*c2=NULL;
    VEC *r = NULL, *z = NULL, *v1 = NULL;
    VEC **v_ptr=NULL, *x;

    assert(b->nvec == 1);
    x = *x_ptr;
    nlocal = x->map->nlocal;
    nglobal = x->map->nglobal;
    /*  allocation of matrix H and Vec v */
    v_ptr=(VEC**)phgAlloc((restart+2)*sizeof(*v_ptr));
    H=(FLOAT **)phgAlloc((restart+2)*sizeof(*H));
    HH=(FLOAT **)phgAlloc((restart+2)*sizeof(*HH));
    h_data=(FLOAT*)phgAlloc((restart+2)*(restart+2)*sizeof(*h_data));
    hh_data=(FLOAT*)phgAlloc((restart+2)*(restart+2)*sizeof(*h_data));
    for(i=0;i<restart+2;i++)
    {
    	v_ptr[i]=phgMapCreateVec(x->map, 1);
	H[i]=h_data+i*(restart+2);	
	HH[i]=hh_data+i*(restart+2);	
    }

    y=(FLOAT *)phgAlloc((restart+2)*4*sizeof(*y));
    s=y+restart+2;
    c1=y+(restart+2)*2;
    c2=y+(restart+2)*3;
    z=phgMapCreateVec(x->map, 1);
     
    /* loop for 'MaxIter' GMRES cycles */
     /* v_ptr[0]= r = b - A * x */
    r=v_ptr[0];
    phgVecCopy(b, &r);
#if USE_MPI
    if (!IsZero(phgVecNorm1(x, 0, NULL) / A->cmap->nglobal))
#else
    if (!IsZero(phgVecNorm1(x, 0, NULL) / A->cmap->nlocal))
#endif
    {
        phgMatVec(MAT_OP_N, -1.0, A, x, 1.0, &r);
    }

    b_norm = phgVecNorm2(b, 0, NULL);
    if (IsZero(b_norm))
       b_norm = 1.0;

    r_norm  = residual = phgVecNorm2(r, 0, NULL);

    tol=rtol*b_norm;
    if(tol < atol)
       tol = atol;
     
    if (phgVerbosity > 0) {
          phgPrintf("=====================================\n");
          phgPrintf("Iters     resid.norm     rel.res.norm\n");
          phgPrintf("-----    ------------    ------------\n");
    }
    nits = 0;

    /* M*z = b - A*x */
    verb = phgVerbosity;
    if (phgVerbosity > 0)
        phgVerbosity--;
    if(pc!=NULL)
        pc->pc_proc(pc->pc_solver, r, &z);
    else
        void_pc(NULL, r, &z);
    phgVerbosity = verb;
  
    while (nits < max_its) {
        v_norm = s[1]= r_norm = phgVecNorm2(z, 0, NULL);
        phgMatVec(MAT_OP_N,1.0/r_norm,NULL,z,0.0,&v_ptr[1]);
        /* GMRES iteration */
        i = 0;
        bzero(h_data,(restart+2)*(restart+2)*sizeof(*h_data));
        bzero(hh_data,(restart+2)*(restart+2)*sizeof(*h_data));
	/* v_ptr[1...i]= (M^-1 * A)^(i-1) * (M^-1)(b-A*x)  */
	while(i < restart && (r_norm > tol || false_conv) &&nits < max_its )
	{
	   i++,nits++;
	   v1=v_ptr[i+1];
	   /* M*v[j+1] = A*v[j] */
	   phgMatVec(MAT_OP_N, 1.0, A, v_ptr[i], 0.0, &r);
	   verb = phgVerbosity;
	   if (phgVerbosity > 0)
		phgVerbosity--;
	   if (pc != NULL)
	        pc->pc_proc(pc->pc_solver, r, &v1);
	   else
	        void_pc(NULL, r, &v1);
	   phgVerbosity = verb;

	   /* modified Gram-Schmidt orthogonalization */
	   for(k=1;k<=i;k++){
		/* H[i][k]=<v[i+1],v[k]> */
		HH[k][i] = H[i][k]=phgVecDot(v1, 0, v_ptr[k], 0, NULL);
		/* v[i+1]=v[i+1]- H[i][k]*v[k] */
		phgMatVec(MAT_OP_N,-H[i][k],NULL,v_ptr[k],1.0,&v1); 
	   } /* end for k-iter */

	   HH[i+1][i] = H[i][i+1]= phgVecNorm2(v1, 0, NULL);
	   if(IsZero(H[i][i+1])) {
		phgInfo(2,"%le happy break i=%d ", (double)H[i][i+1], i);
		i--;
		break_flag=TRUE;
		break ;
	    }
	   phgMatVec(MAT_OP_N,0.0,NULL,NULL,1.0/H[i][i+1],&v1);

	   /* Q-R algoritm */
	   for (j = 1; j < i; j++) {
 	       h1= c1[j] * H[i][j] + c2[j] * H[i][j+1];
               h2 = - c2[j] * H[i][j] + c1[j] * H[i][j+1];
               H[i][j]=h1;
               H[i][j+1]=h2;
            }

	    tmp = Sqrt(H[i][i] * H[i][i] + H[i][i+1] * H[i][i+1]);
	    if(IsZero(tmp)) {tmp=1.e-16;}
            c1[i] = H[i][i] / tmp;
            c2[i] = H[i][i+1] /tmp ;
            s[i+1] = - c2[i] * s[i];
            s[i] = c1[i] * s[i];
	    H[i][i] = c1[i]*H[i][i]+c2[i]* H[i][i+1];
	    r_norm=(double)Fabs(s[i+1]);
	   
            if (phgVerbosity > 0) {
	        phgPrintf("% 5d    %e    %e",nits,r_norm,r_norm/b_norm);
	        if (pc != NULL /*&& pc->pc_solver != solver*/
		    && pc->pc_solver->nits > 0)
		    phgPrintf(" (%d iteration%s in PC)\n", pc->pc_solver->nits,
				pc->pc_solver->nits > 1 ? "s" : "");
	         else
		     phgPrintf("\n");
	    }
        } /* end-for-while(i)-sub-space */

	   /* solve H * y = s */
        for (j = i; j > 0; j--) {
            y[j] = s[j] / H[j][j];
            for (k = j - 1 ; k > 0; k--)
	    {
               s[k] -= H[j][k] * y[j];
	    }
         }
       /* Update x */
        /*  r =  V * y   */
#if 0
	phgMatVec(MAT_OP_N,y[i],NULL,v_ptr[i],0.0,&z);
	for(j=i-1;j>=1;j--){
           phgMatVec(MAT_OP_N,y[j],NULL,v_ptr[j],1.0,&z);
	}
#else
	if(i>0){ 
	    phgMatVec(MAT_OP_N,y[1],NULL,v_ptr[1],0.0,&z);
	    for(j=2;j<=i;j++){
                phgMatVec(MAT_OP_N,y[j],NULL,v_ptr[j],1.0,&z);
	    }
	}
#endif
        /*  x = x +  r  */
   	phgMatVec(MAT_OP_N,1.0,NULL,z,1.0,&x);
	false_conv=FALSE;

        if(r_norm <= tol)
	 {
            /* r = b - A * x */
            phgVecCopy(b, &r);
            phgMatVec(MAT_OP_N, -1.0, A, x, 1.0, &r);
	    residual= phgVecNorm2(r, 0, NULL);
	    if (residual < rtol*b_norm ||residual < atol){
	      if(phgVerbosity>0)
	         phgPrintf("GMRES iteration %d ||b-Ax||_2 = %e  rel.res %e\n",	
				nits, residual,residual/b_norm);
	        break;
            }else if(break_flag){
		     if(phgVerbosity>0){
                          phgPrintf("You reached the happy break down.\n");
	                  phgPrintf("GMRES iteration %d ||b-Ax||_2 = %e  rel.res %e\n",	
				nits, residual,residual/b_norm);
		      }
		      break;
		}else if (i < restart ){
			 if(phgVerbosity>0) {
	                   phgPrintf("false convergence, restart iterative! \n");
                         }
		         false_conv=TRUE;
		}
	
          verb = phgVerbosity;
          if (phgVerbosity > 0)
             phgVerbosity--;
          if(pc!=NULL)
	     pc->pc_proc(pc->pc_solver, r, &z);
	  else
	     void_pc(NULL, r, &z);
	  phgVerbosity = verb;
	  r_norm = phgVecNorm2(z, 0, NULL);
	 }else{
/* computer residual vector and continue loop 
 * z_{i+1}=M^-1 * r_{i+1} =M^-1 * (b  - A*x_{i+1})
 *         = M^-1 *(b  - A(x0 + V_i y))
 *         = z0 - M^-1 *A*V_i* y 
 *         = z0 - V_{i+1} * {HH_i}^T * y
*/	
#if 0
           for (j=1; j <=i+1; j++)
           {
	     s[j]=0.;
	     for(k=1;k<=i;k++)
	       s[j]-=HH[j][k]*y[k];
           }

	   if(i>0){
	       phgMatVec(MAT_OP_N,s[1]+v_norm, NULL,v_ptr[1] , 0.0, &z);
               for (j=2; j <=i+1; j++)
	       phgMatVec(MAT_OP_N,s[j], NULL,v_ptr[j] , 1.0, &z);
           }
#else
           phgVecCopy(b, &r);
           phgMatVec(MAT_OP_N, -1.0, A, x, 1.0, &r);
           verb = phgVerbosity;
           if (phgVerbosity > 0)
               phgVerbosity--;
           if(pc!=NULL)
	       pc->pc_proc(pc->pc_solver, r, &z);
	   else
	       void_pc(NULL, r, &z);
	   phgVerbosity = verb;
#endif
        }
    }/* main-loop */
    
    if (res != NULL)
      *res = residual;

    phgVecDestroy(&z);

    for(i=0;i<restart+2;i++)
	phgVecDestroy(&v_ptr[i]);
    phgFree(v_ptr);
    phgFree(H);
    phgFree(HH);
    phgFree(h_data);
    phgFree(hh_data);
    phgFree(y);

    return nits;
}

static int
Init(SOLVER *solver)
{
    phgInfo(2, "Create %s solver\n", solver->oem_solver->name);
    solver->oem_data = phgCalloc(1, sizeof(OEM_DATA));

    if (solver->oem_solver == SOLVER_PCG) {
	_t->pc_id = pcg_pc_id;
    }
    else {
	_t->pc_id = gmres_pc_id;
	_t->gmres_restart = gmres_restart;
    }

    return 0;
}

static int
Create(SOLVER *solver)
{
    if (solver->pc == NULL)
	phgSolverSetPC(solver,solver,pc_list[_t->pc_id]);

    _t->pc = (SOLVER_PC*) solver->pc;

    return 0;
}

static int
Destroy(SOLVER *solver)
{
    if (solver->oem_data == NULL)
	return 0;

    phgInfo(2, "Destroy PCG Solver\n");
    phgFree(solver->oem_data);

    return 0;
}

#define Assemble		NULL
#define AddMatrixEntries	NULL
#define AddRHSEntries		NULL

static int
PCG_Solve(SOLVER *solver, VEC *x, BOOLEAN destroy)
{
    INT nits;
    FLOAT residual = 1.0;
    VEC *B;

    B = phgMapCreateVec(x->map, 0);
    B->data = solver->rhs->data;
    B->offp_data = solver->rhs->offp_data;
    assert(solver->rhs->assembled);
    B->nvec = 1;
    nits = phgSolverPCG__(solver->mat, B, &x, _t->pc, solver->maxit,
			  solver->rtol, solver->atol, &residual, &_t->warned);
    B->data = NULL;
    B->offp_data = NULL;
    if (destroy) {
	phgMatDestroy(&solver->mat);
	solver->sent_to_oem = TRUE;
    }
    phgVecDestroy(&B);

    solver->residual = residual;
    return solver->nits = nits;
}

static int
GMRES_Solve(SOLVER *solver, VEC *x, BOOLEAN destroy)
{
    INT nits;
    FLOAT residual = 1.0;
    VEC *B;

    B = phgMapCreateVec(x->map, 0);
    B->data = solver->rhs->data;
    B->offp_data = solver->rhs->offp_data;
    assert(solver->rhs->assembled);
    B->nvec = 1;
    nits = phgSolverGMRES(solver->mat, B, &x, _t->pc, _t->gmres_restart,
			  solver->maxit, solver->rtol, &residual);
    B->data = NULL;
    B->offp_data = NULL;
    if (destroy) {
	phgMatDestroy(&solver->mat);
	solver->sent_to_oem = TRUE;
    }
    phgVecDestroy(&B);

    solver->residual = residual;
    return solver->nits = nits;
}

/*--------------------------------------------------------------------*/

OEM_SOLVER phgSolverPCG_ = {
    "PCG", PCGRegisterOptions,
    Initialize, Finalize, Init, Create, Destroy, AddMatrixEntries,
    AddRHSEntries, Assemble, PCG_Solve
};

OEM_SOLVER phgSolverGMRES_ = {
    "GMRES", GMRESRegisterOptions,
    Initialize, Finalize, Init, Create, Destroy, AddMatrixEntries,
    AddRHSEntries, Assemble, GMRES_Solve
};
